"""
Test the win_wua state module
"""

from collections import namedtuple

import pytest

import salt.states.win_wua as win_wua
import salt.utils.platform
import salt.utils.win_update as win_update
from tests.support.mock import MagicMock, patch


@pytest.fixture
def updates_list():
    updated_list = {
        "ca3bb521-a8ea-4e26-a563-2ad6e3108b9a": {
            "KBs": ["KB4481252"],
            "Installed": False,
            "Title": "Blank",
        },
        "07609d43-d518-4e77-856e-d1b316d1b8a8": {
            "KBs": ["KB925673"],
            "Installed": False,
            "Title": "Blank",
        },
        "fbaa5360-a440-49d8-a3b6-0c4fc7ecaa19": {
            "KBs": ["KB4481252"],
            "Installed": False,
            "Title": "Blank",
        },
        "a873372b-7a5c-443c-8022-cd59a550bef4": {
            "KBs": ["KB3193497"],
            "Installed": False,
            "Title": "Blank",
        },
        "14075cbe-822e-4004-963b-f50e08d45563": {
            "KBs": ["KB4540723"],
            "Installed": False,
            "Title": "Blank",
        },
        "d931e99c-4dda-4d39-9905-0f6a73f7195f": {
            "KBs": ["KB3193498"],
            "Installed": False,
            "Title": "Blank",
        },
        "afda9e11-44a0-4602-9e9b-423af11ecaed": {
            "KBs": ["KB4541329"],
            "Installed": False,
            "Title": "Blank",
        },
        "a0f997b1-1abe-4a46-941f-b37f732f9fbd": {
            "KBs": ["KB3193497"],
            "Installed": False,
            "Title": "Blank",
        },
        "eac02b09-d745-4891-b80f-400e0e5e4b6d": {
            "KBs": ["KB4052623"],
            "Installed": False,
            "Title": "KB4052623: Really long title that exceeds 40 characters",
        },
        "0689e74b-54d1-4f55-a916-96e3c737db90": {
            "KBs": ["KB890830"],
            "Installed": False,
            "Title": "Blank",
        },
    }
    return updated_list


@pytest.fixture
def updates_list_none():
    updates_list_empty = {}
    return updates_list_empty


@pytest.fixture
def updates_summary():
    updates_summary_d = {"Installed": 10}
    return updates_summary_d


class Updates:
    def __init__(self):
        self.updates = []


@pytest.fixture
def configure_loader_modules():
    return {win_wua: {"__opts__": {"test": False}, "__env__": "base"}}


@pytest.fixture
def update_records():
    return namedtuple(
        "UpdateRecord",
        ["KBArticleIDs", "Identity", "IsDownloaded", "IsInstalled", "Title"],
    )


@pytest.fixture
def update_records_identity():
    return namedtuple("UpdateRecordIdentity", "UpdateID")


@pytest.mark.skip_unless_on_windows
def test_uptodate_no_updates(updates_list_none):
    """
    Test uptodate function with no updates found.
    """
    expected = {
        "name": "NA",
        "changes": {},
        "result": True,
        "comment": "No updates found",
    }

    class NoUpdates(Updates):
        @staticmethod
        def updates():  # pylint: disable=method-hidden
            return updates_list_none

        @staticmethod
        def count():
            return len(updates_list_none)

    patch_winapi_com = patch("salt.utils.winapi.Com", autospec=True)
    patch_win32com = patch("win32com.client.Dispatch", autospec=True)
    patch_win_update_agent = patch.object(
        salt.utils.win_update.WindowsUpdateAgent, "refresh", autospec=True
    )
    patch_win_update = patch.object(
        salt.utils.win_update, "Updates", autospec=True, return_value=NoUpdates()
    )

    with patch_winapi_com, patch_win32com, patch_win_update_agent, patch_win_update:
        result = win_wua.uptodate(name="NA")
        assert result == expected


@pytest.mark.skip_unless_on_windows
def test_uptodate_test_mode():
    """
    Test uptodate function in test=true mode.
    """
    expected = {
        "name": "NA",
        "changes": {},
        "result": None,
        "comment": "Updates will be installed:",
    }
    patch_winapi_com = patch("salt.utils.winapi.Com", autospec=True)
    patch_win32com = patch("win32com.client.Dispatch", autospec=True)
    patch_win_update_agent = patch.object(
        salt.utils.win_update.WindowsUpdateAgent, "refresh", autospec=True
    )
    patch_opts = patch.dict(win_wua.__opts__, {"test": True})

    with patch_winapi_com, patch_win32com, patch_win_update_agent, patch_opts:
        wua = win_update.WindowsUpdateAgent(online=False)
        wua._updates = [MagicMock(IsInstalled=False, IsDownloaded=False)]
        result = win_wua.uptodate(name="NA")
        assert result == expected


@pytest.mark.skip_unless_on_windows
def test_uptodate(updates_list):
    """
    Test uptodate function with some updates found.
    """
    expected = {
        "name": "NA",
        "changes": {
            "failed": {
                "a0f997b1-1abe-4a46-941f-b37f732f9fbd": {
                    "KBs": ["KB3193497"],
                    "Title": "Blank",
                },
                "afda9e11-44a0-4602-9e9b-423af11ecaed": {
                    "KBs": ["KB4541329"],
                    "Title": "Blank",
                },
                "eac02b09-d745-4891-b80f-400e0e5e4b6d": {
                    "KBs": ["KB4052623"],
                    "Title": "KB4052623: Really long title that exceeds 40 characters",
                },
            },
            "superseded": {
                "eac02c07-d744-4892-b80f-312d045e4ccc": {
                    "KBs": ["KB4052444"],
                    "Title": "Superseded Update",
                }
            },
        },
        "result": False,
        "comment": "Some updates failed to install\nSome updates were superseded",
    }

    updates_not_installed = {
        "afda9e11-44a0-4602-9e9b-423af11ecaed": {
            "KBs": ["KB4541329"],
            "Installed": False,
            "Title": "Blank",
        },
        "a0f997b1-1abe-4a46-941f-b37f732f9fbd": {
            "KBs": ["KB3193497"],
            "Installed": False,
            "Title": "Blank",
        },
        "eac02b09-d745-4891-b80f-400e0e5e4b6d": {
            "KBs": ["KB4052623"],
            "Installed": False,
            "Title": "KB4052623: Really long title that exceeds 40 characters",
        },
        "eac02c07-d744-4892-b80f-312d045e4ccc": {
            "KBs": ["KB4052444"],
            "Installed": False,
            "Title": "Superseded Update",
        },
    }

    fake_updates = MagicMock()
    fake_updates.list.return_value = updates_not_installed

    patch_win_wua_update = patch(
        "salt.utils.win_update.Updates",
        autospec=True,
        return_value=fake_updates,
    )

    fake_wua_updates = MagicMock()
    fake_wua_updates.list.return_value = updates_list

    fake_wua = MagicMock()
    fake_wua.updates.return_value = fake_wua_updates
    patch_wua = patch(
        "salt.utils.win_update.WindowsUpdateAgent",
        autospec=True,
        return_value=fake_wua,
    )

    patch_winapi_com = patch("salt.utils.winapi.Com", autospec=True)
    patch_win32 = patch("win32com.client.Dispatch", autospec=True)
    patch_opts = patch.dict(win_wua.__opts__, {"test": False})

    with patch_winapi_com, patch_win32, patch_wua, patch_win_wua_update, patch_opts:
        result = win_wua.uptodate(name="NA")
        assert result == expected


@pytest.mark.skip_unless_on_windows
def test_installed(update_records, update_records_identity):
    """
    Test installed function
    """

    update_search_obj = {
        update_records(
            KBArticleIDs=("4052623",),
            Identity=update_records_identity(
                UpdateID="eac02b09-d745-4891-b80f-400e0e5e4b6d"
            ),
            IsDownloaded=False,
            IsInstalled=False,
            Title="KB4052623: Really long title that exceeds 40 characters",
        ),
    }

    update_search_dict = {
        "eac02b09-d745-4891-b80f-400e0e5e4b6d": {
            "Downloaded": True,
            "KBs": ["KB4052623"],
            "Installed": True,
            "NeedsReboot": True,
            "Title": "KB4052623: Really long title that exceeds 40 characters",
        },
    }

    updates_refresh = {
        "a0f997b1-1abe-4a46-941f-b37f732f9fbd": {
            "Downloaded": False,
            "KBs": ["KB3193497"],
            "Installed": False,
            "NeedsReboot": False,
            "Title": "Update 1",
        },
        "eac02b09-d745-4891-b80f-400e0e5e4b6d": {
            "Downloaded": True,
            "KBs": ["KB4052623"],
            "Installed": True,
            "NeedsReboot": True,
            "Title": "KB4052623: Really long title that exceeds 40 characters",
        },
        "eac02c07-d744-4892-b80f-312d045e4ccc": {
            "Downloaded": True,
            "KBs": ["KB4052444"],
            "Installed": True,
            "NeedsReboot": False,
            "Title": "Update 3",
        },
    }

    # Mocks the connection to the Windows Update Agent
    mock_wua = MagicMock()
    # Mocks the initial search
    mock_wua.search = MagicMock()
    # Mocks the number of updates found.
    mock_wua.search().count.return_value = 1
    # Mocks the the updates collection object
    mock_wua.search().updates = update_search_obj

    # This mocks the updates collection in the install variable. This will
    # get populated to as matches are found with the Add method
    mock_updates = MagicMock()
    # Needs to return the number of updates that need to be installed
    # (IsInstalled = False)
    mock_updates.count.return_value = 1
    # Returns the updates that need to be installed as a dict
    mock_updates.list.return_value = update_search_dict

    # This gives us post_info
    mock_wua.updates = MagicMock()
    # Mock a refresh of the updates recognized by the machine. This would
    # occur post install. This is compared with the updates on the machine
    # to determine if the update was successful
    mock_wua.updates().list.return_value = updates_refresh

    patch_winapi_com = patch("salt.utils.winapi.Com", autospec=True)
    patch_dispatch = patch("win32com.client.Dispatch", autospec=True)
    patch_wua = patch(
        "salt.utils.win_update.WindowsUpdateAgent",
        autospec=True,
        return_value=mock_wua,
    )
    patch_update_collection = patch(
        "salt.utils.win_update.Updates", autospec=True, return_value=mock_updates
    )
    patch_opts = patch.dict(win_wua.__opts__, {"test": False})

    with (
        patch_winapi_com
    ), patch_dispatch, patch_wua, patch_update_collection, patch_opts:
        expected = {
            "changes": {
                "installed": {
                    "eac02b09-d745-4891-b80f-400e0e5e4b6d": {
                        "KBs": ["KB4052623"],
                        "NeedsReboot": True,
                        "Title": "KB4052623: Really long title that exceeds 40 characters",
                    }
                }
            },
            "comment": "Updates installed successfully",
            "name": "KB4062623",
            "result": True,
        }
        result = win_wua.installed(name="KB4062623")
        assert result == expected


@pytest.mark.skip_unless_on_windows
def test_installed_no_updates():
    """
    Test installed function when no updates are found.
    """
    # Mocks the connection to the Windows Update Agent
    mock_wua = MagicMock()
    # Mocks the initial search
    mock_wua.search = MagicMock()
    # Mocks the number of updates found.
    mock_wua.search().count.return_value = 0

    patch_winapi_com = patch("salt.utils.winapi.Com", autospec=True)
    patch_dispatch = patch("win32com.client.Dispatch", autospec=True)
    patch_wua = patch(
        "salt.utils.win_update.WindowsUpdateAgent",
        autospec=True,
        return_value=mock_wua,
    )

    with patch_winapi_com, patch_dispatch, patch_wua:
        expected = {
            "name": "KB4062623",
            "changes": {},
            "result": True,
            "comment": "No updates found",
        }
        result = win_wua.installed(name="KB4062623")
        assert result == expected


@pytest.mark.skip_unless_on_windows
def test_installed_test_mode(update_records, update_records_identity):
    """
    Test installed function in test mode
    """

    update_search_obj = {
        update_records(
            KBArticleIDs=("4052623",),
            Identity=update_records_identity(
                UpdateID="eac02b09-d745-4891-b80f-400e0e5e4b6d"
            ),
            IsDownloaded=False,
            IsInstalled=False,
            Title="Update 2",
        ),
    }

    # Mocks the connection to the Windows Update Agent
    mock_wua = MagicMock()
    # Mocks the initial search
    mock_wua.search = MagicMock()
    # Mocks the number of updates found.
    mock_wua.search().count.return_value = 1
    # Mocks the the updates collection object
    mock_wua.search().updates = update_search_obj

    # This mocks the updates collection in the install variable. This will
    # get populated to as matches are found with the Add method
    mock_updates = MagicMock()
    # Needs to return the number of updates that need to be installed
    # (IsInstalled = False)
    mock_updates.count.return_value = 1

    patch_winapi_com = patch("salt.utils.winapi.Com", autospec=True)
    patch_dispatch = patch("win32com.client.Dispatch", autospec=True)
    patch_wua = patch(
        "salt.utils.win_update.WindowsUpdateAgent",
        autospec=True,
        return_value=mock_wua,
    )
    patch_update_collection = patch(
        "salt.utils.win_update.Updates", autospec=True, return_value=mock_updates
    )
    patch_opts = patch.dict(win_wua.__opts__, {"test": True})

    with (
        patch_winapi_com
    ), patch_dispatch, patch_wua, patch_update_collection, patch_opts:
        expected = {
            "changes": {},
            "comment": "Updates will be installed:",
            # I don't know how to mock this part so the list will show up.
            # It's an update collection object populated using the Add
            # method. But this works for now
            "name": "KB4062623",
            "result": None,
        }
        result = win_wua.installed(name="KB4062623")
        assert result == expected


@pytest.mark.skip_unless_on_windows
def test_installed_already_installed(update_records, update_records_identity):
    """
    Test installed function when the update is already installed
    """

    update_search_obj = {
        update_records(
            KBArticleIDs=("4052623",),
            Identity=update_records_identity(
                UpdateID="eac02b09-d745-4891-b80f-400e0e5e4b6d"
            ),
            IsDownloaded=True,
            IsInstalled=True,
            Title="Update 2",
        ),
    }

    # Mocks the connection to the Windows Update Agent
    mock_wua = MagicMock()
    # Mocks the initial search
    mock_wua.search = MagicMock()
    # Mocks the number of updates found.
    mock_wua.search().count.return_value = 1
    # Mocks the the updates collection object
    mock_wua.search().updates = update_search_obj

    # This mocks the updates collection in the install variable. This will
    # get populated to as matches are found with the Add method
    mock_updates = MagicMock()
    # Needs to return the number of updates that need to be installed
    # (IsInstalled = False)
    mock_updates.count.return_value = 0

    patch_winapi_com = patch("salt.utils.winapi.Com", autospec=True)
    patch_dispatch = patch("win32com.client.Dispatch", autospec=True)
    patch_wua = patch(
        "salt.utils.win_update.WindowsUpdateAgent",
        autospec=True,
        return_value=mock_wua,
    )
    patch_update_collection = patch(
        "salt.utils.win_update.Updates", autospec=True, return_value=mock_updates
    )
    patch_opts = patch.dict(win_wua.__opts__, {"test": True})

    with (
        patch_winapi_com
    ), patch_dispatch, patch_wua, patch_update_collection, patch_opts:
        expected = {
            "changes": {},
            "comment": "Updates already installed: KB4052623",
            "name": "KB4062623",
            "result": True,
        }
        result = win_wua.installed(name="KB4062623")
        assert result == expected


@pytest.mark.skip_unless_on_windows
def test_removed(update_records, update_records_identity):
    """
    Test removed function
    """

    update_search_obj = {
        update_records(
            KBArticleIDs=("4052623",),
            Identity=update_records_identity(
                UpdateID="eac02b09-d745-4891-b80f-400e0e5e4b6d"
            ),
            IsDownloaded=False,
            IsInstalled=True,
            Title="KB4052623: Really long title that exceeds 40 characters",
        ),
    }

    update_search_dict = {
        "eac02b09-d745-4891-b80f-400e0e5e4b6d": {
            "Downloaded": True,
            "KBs": ["KB4052623"],
            "Installed": False,
            "NeedsReboot": True,
            "Title": "KB4052623: Really long title that exceeds 40 characters",
        },
    }

    updates_refresh = {
        "a0f997b1-1abe-4a46-941f-b37f732f9fbd": {
            "Downloaded": False,
            "KBs": ["KB3193497"],
            "Installed": False,
            "NeedsReboot": False,
            "Title": "Update 1",
        },
        "eac02b09-d745-4891-b80f-400e0e5e4b6d": {
            "Downloaded": True,
            "KBs": ["KB4052623"],
            "Installed": False,
            "NeedsReboot": True,
            "Title": "KB4052623: Really long title that exceeds 40 characters",
        },
        "eac02c07-d744-4892-b80f-312d045e4ccc": {
            "Downloaded": True,
            "KBs": ["KB4052444"],
            "Installed": True,
            "NeedsReboot": False,
            "Title": "Update 3",
        },
    }

    # Mocks the connection to the Windows Update Agent
    mock_wua = MagicMock()
    # Mocks the initial search
    mock_wua.search = MagicMock()
    # Mocks the number of updates found.
    mock_wua.search().count.return_value = 1
    # Mocks the the updates collection object
    mock_wua.search().updates = update_search_obj

    # This mocks the updates collection in the uninstall variable. This will
    # get populated as matches are found with the Add method
    mock_updates = MagicMock()
    # Needs to return the number of updates that need to be installed
    # (IsInstalled = False)
    mock_updates.count.return_value = 1
    # Returns the updates that need to be installed as a dict
    mock_updates.list.return_value = update_search_dict

    # This gives us post_info
    mock_wua.updates = MagicMock()
    # Mock a refresh of the updates recognized by the machine. This would
    # occur post uninstall. This is compared with the updates on the machine
    # to determine if the removal was successful
    mock_wua.updates().list.return_value = updates_refresh

    patch_winapi_com = patch("salt.utils.winapi.Com", autospec=True)
    patch_dispatch = patch("win32com.client.Dispatch", autospec=True)
    patch_wua = patch(
        "salt.utils.win_update.WindowsUpdateAgent",
        autospec=True,
        return_value=mock_wua,
    )
    patch_update_collection = patch(
        "salt.utils.win_update.Updates", autospec=True, return_value=mock_updates
    )
    patch_opts = patch.dict(win_wua.__opts__, {"test": False})

    with (
        patch_winapi_com
    ), patch_dispatch, patch_wua, patch_update_collection, patch_opts:
        expected = {
            "changes": {
                "removed": {
                    "eac02b09-d745-4891-b80f-400e0e5e4b6d": {
                        "KBs": ["KB4052623"],
                        "NeedsReboot": True,
                        "Title": "KB4052623: Really long title that exceeds 40 characters",
                    }
                }
            },
            "comment": "Updates removed successfully",
            "name": "KB4062623",
            "result": True,
        }
        result = win_wua.removed(name="KB4062623")
        assert result == expected


@pytest.mark.skip_unless_on_windows
def test_removed_test_mode(update_records, update_records_identity):
    """
    Test removed function in test mode
    """

    update_search_obj = {
        update_records(
            KBArticleIDs=("4052623",),
            Identity=update_records_identity(
                UpdateID="eac02b09-d745-4891-b80f-400e0e5e4b6d"
            ),
            IsDownloaded=False,
            IsInstalled=True,
            Title="KB4052623: Really long title that exceeds 40 characters",
        ),
    }

    # Mocks the connection to the Windows Update Agent
    mock_wua = MagicMock()
    # Mocks the initial search
    mock_wua.search = MagicMock()
    # Mocks the number of updates found.
    mock_wua.search().count.return_value = 1
    # Mocks the the updates collection object
    mock_wua.search().updates = update_search_obj

    # This mocks the updates collection in the install variable. This will
    # get populated to as matches are found with the Add method
    mock_updates = MagicMock()
    # Needs to return the number of updates that need to be installed
    # (IsInstalled = False)
    mock_updates.count.return_value = 1

    patch_winapi_com = patch("salt.utils.winapi.Com", autospec=True)
    patch_dispatch = patch("win32com.client.Dispatch", autospec=True)
    patch_wua = patch(
        "salt.utils.win_update.WindowsUpdateAgent",
        autospec=True,
        return_value=mock_wua,
    )
    patch_update_collection = patch(
        "salt.utils.win_update.Updates", autospec=True, return_value=mock_updates
    )
    patch_opts = patch.dict(win_wua.__opts__, {"test": True})

    with (
        patch_winapi_com
    ), patch_dispatch, patch_wua, patch_update_collection, patch_opts:
        expected = {
            "changes": {},
            "comment": "Updates will be removed:",
            # I don't know how to mock this part so the list will show up.
            # It's an update collection object populated using the Add
            # method. But this works for now
            "name": "KB4062623",
            "result": None,
        }
        result = win_wua.removed(name="KB4062623")
        assert result == expected



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.mark.parametrize("test", (False, True))
def test_uncomment(file, tmp_path, test):
    """
    file.uncomment
    """
    name = tmp_path / "testfile"
    # write a line to file
    name.write_text("#comment_me")

    ret = file.uncomment(name=str(name), regex="^comment", test=test)
    if test is True:
        assert ret.result is None
        assert "#comment" in name.read_text()
    else:
        assert ret.result is True
        assert "#comment" not in name.read_text()



"""
    tests.unit.xmlutil_test
    ~~~~~~~~~~~~~~~~~~~~
"""

import xml.etree.ElementTree as ET

import salt.utils.xmlutil as xml
from tests.support.unit import TestCase


class XMLUtilTestCase(TestCase):
    """
    Tests that salt.utils.xmlutil properly parses XML data and returns as a properly formatted
    dictionary. The default method of parsing will ignore attributes and return only the child
    items. The full method will include parsing attributes.
    """

    def setUp(self):

        # Populate our use cases for specific XML formats.
        self.cases = {
            "a": {
                "xml": "<parent>data</parent>",
                "legacy": {"parent": "data"},
                "full": "data",
            },
            "b": {
                "xml": '<parent value="data">data</parent>',
                "legacy": {"parent": "data"},
                "full": {"parent": "data", "value": "data"},
            },
            "c": {
                "xml": (
                    '<parent><child>data</child><child value="data">data</child>'
                    '<child value="data"/><child/></parent>'
                ),
                "legacy": {
                    "child": [
                        "data",
                        {"child": "data"},
                        {"child": None},
                        {"child": None},
                    ]
                },
                "full": {
                    "child": [
                        "data",
                        {"child": "data", "value": "data"},
                        {"value": "data"},
                        None,
                    ]
                },
            },
            "d": {
                "xml": (
                    '<parent value="data" another="data"><child>data</child></parent>'
                ),
                "legacy": {"child": "data"},
                "full": {"child": "data", "another": "data", "value": "data"},
            },
            "e": {
                "xml": (
                    '<parent value="data" another="data"><child'
                    ' value="data">data</child></parent>'
                ),
                "legacy": {"child": "data"},
                "full": {
                    "child": {"child": "data", "value": "data"},
                    "another": "data",
                    "value": "data",
                },
            },
            "f": {
                "xml": (
                    '<parent><child><sub-child value="data">data</sub-child></child>'
                    "<child>data</child></parent>"
                ),
                "legacy": {"child": [{"sub-child": "data"}, {"child": "data"}]},
                "full": {
                    "child": [
                        {"sub-child": {"value": "data", "sub-child": "data"}},
                        "data",
                    ]
                },
            },
        }

    def test_xml_case_a(self):
        xmldata = ET.fromstring(self.cases["a"]["xml"])
        defaultdict = xml.to_dict(xmldata)
        self.assertEqual(defaultdict, self.cases["a"]["legacy"])

    def test_xml_case_a_legacy(self):
        xmldata = ET.fromstring(self.cases["a"]["xml"])
        defaultdict = xml.to_dict(xmldata, False)
        self.assertEqual(defaultdict, self.cases["a"]["legacy"])

    def test_xml_case_a_full(self):
        xmldata = ET.fromstring(self.cases["a"]["xml"])
        defaultdict = xml.to_dict(xmldata, True)
        self.assertEqual(defaultdict, self.cases["a"]["full"])

    def test_xml_case_b(self):
        xmldata = ET.fromstring(self.cases["b"]["xml"])
        defaultdict = xml.to_dict(xmldata)
        self.assertEqual(defaultdict, self.cases["b"]["legacy"])

    def test_xml_case_b_legacy(self):
        xmldata = ET.fromstring(self.cases["b"]["xml"])
        defaultdict = xml.to_dict(xmldata, False)
        self.assertEqual(defaultdict, self.cases["b"]["legacy"])

    def test_xml_case_b_full(self):
        xmldata = ET.fromstring(self.cases["b"]["xml"])
        defaultdict = xml.to_dict(xmldata, True)
        self.assertEqual(defaultdict, self.cases["b"]["full"])

    def test_xml_case_c(self):
        xmldata = ET.fromstring(self.cases["c"]["xml"])
        defaultdict = xml.to_dict(xmldata)
        self.assertEqual(defaultdict, self.cases["c"]["legacy"])

    def test_xml_case_c_legacy(self):
        xmldata = ET.fromstring(self.cases["c"]["xml"])
        defaultdict = xml.to_dict(xmldata, False)
        self.assertEqual(defaultdict, self.cases["c"]["legacy"])

    def test_xml_case_c_full(self):
        xmldata = ET.fromstring(self.cases["c"]["xml"])
        defaultdict = xml.to_dict(xmldata, True)
        self.assertEqual(defaultdict, self.cases["c"]["full"])

    def test_xml_case_d(self):
        xmldata = ET.fromstring(self.cases["d"]["xml"])
        defaultdict = xml.to_dict(xmldata)
        self.assertEqual(defaultdict, self.cases["d"]["legacy"])

    def test_xml_case_d_legacy(self):
        xmldata = ET.fromstring(self.cases["d"]["xml"])
        defaultdict = xml.to_dict(xmldata, False)
        self.assertEqual(defaultdict, self.cases["d"]["legacy"])

    def test_xml_case_d_full(self):
        xmldata = ET.fromstring(self.cases["d"]["xml"])
        defaultdict = xml.to_dict(xmldata, True)
        self.assertEqual(defaultdict, self.cases["d"]["full"])

    def test_xml_case_e(self):
        xmldata = ET.fromstring(self.cases["e"]["xml"])
        defaultdict = xml.to_dict(xmldata)
        self.assertEqual(defaultdict, self.cases["e"]["legacy"])

    def test_xml_case_e_legacy(self):
        xmldata = ET.fromstring(self.cases["e"]["xml"])
        defaultdict = xml.to_dict(xmldata, False)
        self.assertEqual(defaultdict, self.cases["e"]["legacy"])

    def test_xml_case_e_full(self):
        xmldata = ET.fromstring(self.cases["e"]["xml"])
        defaultdict = xml.to_dict(xmldata, True)
        self.assertEqual(defaultdict, self.cases["e"]["full"])

    def test_xml_case_f(self):
        xmldata = ET.fromstring(self.cases["f"]["xml"])
        defaultdict = xml.to_dict(xmldata)
        self.assertEqual(defaultdict, self.cases["f"]["legacy"])

    def test_xml_case_f_legacy(self):
        xmldata = ET.fromstring(self.cases["f"]["xml"])
        defaultdict = xml.to_dict(xmldata, False)
        self.assertEqual(defaultdict, self.cases["f"]["legacy"])

    def test_xml_case_f_full(self):
        xmldata = ET.fromstring(self.cases["f"]["xml"])
        defaultdict = xml.to_dict(xmldata, True)
        self.assertEqual(defaultdict, self.cases["f"]["full"])



"""
Execution module for Amazon Route53 written against Boto 3

.. versionadded:: 2017.7.0

:configuration: This module accepts explicit route53 credentials but can also
    utilize IAM roles assigned to the instance through Instance Profiles.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary. More Information available at:

    .. code-block:: yaml

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        route53.keyid: GKTADJGHEIQSXMKKRBJ08H
        route53.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        route53.region: us-east-1

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
          keyid: GKTADJGHEIQSXMKKRBJ08H
          key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
          region: us-east-1

    Note that Route53 essentially ignores all (valid) settings for 'region',
    since there is only one Endpoint (in us-east-1 if you care) and any (valid)
    region setting will just send you there.  It is entirely safe to set it to
    None as well.

:depends: boto3
"""

# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602,W0106


import logging
import re
import time

import salt.utils.compat
import salt.utils.versions
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

try:
    # pylint: disable=unused-import
    import boto3

    # pylint: enable=unused-import
    from botocore.exceptions import ClientError

    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False


def __virtual__():
    """
    Only load if boto libraries exist and if boto libraries are greater than
    a given version.
    """
    return salt.utils.versions.check_boto_reqs()


def __init__(opts):
    if HAS_BOTO3:
        __utils__["boto3.assign_funcs"](__name__, "route53")


def _collect_results(func, item, args, marker="Marker", nextmarker="NextMarker"):
    ret = []
    Marker = args.get(marker, "")
    tries = 10
    while Marker is not None:
        try:
            r = func(**args)
        except ClientError as e:
            if tries and e.response.get("Error", {}).get("Code") == "Throttling":
                # Rate limited - retry
                log.debug("Throttled by AWS API.")
                time.sleep(3)
                tries -= 1
                continue
            log.error("Could not collect results from %s(): %s", func, e)
            return []
        i = r.get(item, []) if item else r
        i.pop("ResponseMetadata", None) if isinstance(i, dict) else None
        ret += i if isinstance(i, list) else [i]
        Marker = r.get(nextmarker)
        args.update({marker: Marker})
    return ret


def _wait_for_sync(change, conn, tries=10, sleep=20):
    for retry in range(1, tries + 1):
        log.info("Getting route53 status (attempt %s)", retry)
        status = "wait"
        try:
            status = conn.get_change(Id=change)["ChangeInfo"]["Status"]
        except ClientError as e:
            if e.response.get("Error", {}).get("Code") == "Throttling":
                log.debug("Throttled by AWS API.")
            else:
                raise
        if status == "INSYNC":
            return True
        time.sleep(sleep)
    log.error("Timed out waiting for Route53 INSYNC status.")
    return False


def find_hosted_zone(
    Id=None,
    Name=None,
    PrivateZone=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Find a hosted zone with the given characteristics.

    Id
        The unique Zone Identifier for the Hosted Zone.  Exclusive with Name.

    Name
        The domain name associated with the Hosted Zone.  Exclusive with Id.
        Note this has the potential to match more then one hosted zone (e.g. a public and a private
        if both exist) which will raise an error unless PrivateZone has also been passed in order
        split the different.

    PrivateZone
        Boolean - Set to True if searching for a private hosted zone.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        Dict, or pillar key pointing to a dict, containing AWS region/key/keyid.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.find_hosted_zone Name=salt.org. \
                profile='{"region": "us-east-1", "keyid": "A12345678AB", "key": "xblahblahblah"}'
    """
    if not _exactly_one((Id, Name)):
        raise SaltInvocationError("Exactly one of either Id or Name is required.")
    if PrivateZone is not None and not isinstance(PrivateZone, bool):
        raise SaltInvocationError(
            "If set, PrivateZone must be a bool (e.g. True / False)."
        )
    if Id:
        ret = get_hosted_zone(Id, region=region, key=key, keyid=keyid, profile=profile)
    else:
        ret = get_hosted_zones_by_domain(
            Name, region=region, key=key, keyid=keyid, profile=profile
        )
    if PrivateZone is not None:
        ret = [
            m for m in ret if m["HostedZone"]["Config"]["PrivateZone"] is PrivateZone
        ]
    if len(ret) > 1:
        log.error(
            "Request matched more than one Hosted Zone (%s). Refine your "
            "criteria and try again.",
            [z["HostedZone"]["Id"] for z in ret],
        )
        ret = []
    return ret


def get_hosted_zone(Id, region=None, key=None, keyid=None, profile=None):
    """
    Return detailed info about the given zone.

    Id
        The unique Zone Identifier for the Hosted Zone.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        Dict, or pillar key pointing to a dict, containing AWS region/key/keyid.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.get_hosted_zone Z1234567690 \
                profile='{"region": "us-east-1", "keyid": "A12345678AB", "key": "xblahblahblah"}'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    args = {"Id": Id}
    return _collect_results(conn.get_hosted_zone, None, args)


def get_hosted_zones_by_domain(Name, region=None, key=None, keyid=None, profile=None):
    """
    Find any zones with the given domain name and return detailed info about them.
    Note that this can return multiple Route53 zones, since a domain name can be used in
    both public and private zones.

    Name
        The domain name associated with the Hosted Zone(s).

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        Dict, or pillar key pointing to a dict, containing AWS region/key/keyid.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.get_hosted_zones_by_domain salt.org. \
                profile='{"region": "us-east-1", "keyid": "A12345678AB", "key": "xblahblahblah"}'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    zones = [
        z
        for z in _collect_results(conn.list_hosted_zones, "HostedZones", {})
        if z["Name"] == _aws_encode(Name)
    ]
    ret = []
    for z in zones:
        ret += get_hosted_zone(
            Id=z["Id"], region=region, key=key, keyid=keyid, profile=profile
        )
    return ret


def list_hosted_zones(
    DelegationSetId=None, region=None, key=None, keyid=None, profile=None
):
    """
    Return detailed info about all zones in the bound account.

    DelegationSetId
        If you're using reusable delegation sets and you want to list all of the hosted zones that
        are associated with a reusable delegation set, specify the ID of that delegation set.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        Dict, or pillar key pointing to a dict, containing AWS region/key/keyid.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.describe_hosted_zones \
                profile='{"region": "us-east-1", "keyid": "A12345678AB", "key": "xblahblahblah"}'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    args = {"DelegationSetId": DelegationSetId} if DelegationSetId else {}
    return _collect_results(conn.list_hosted_zones, "HostedZones", args)


def create_hosted_zone(
    Name,
    VPCId=None,
    VPCName=None,
    VPCRegion=None,
    CallerReference=None,
    Comment="",
    PrivateZone=False,
    DelegationSetId=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create a new Route53 Hosted Zone. Returns a Python data structure with information about the
    newly created Hosted Zone.

    Name
        The name of the domain. This should be a fully-specified domain, and should terminate with
        a period. This is the name you have registered with your DNS registrar. It is also the name
        you will delegate from your registrar to the Amazon Route 53 delegation servers returned in
        response to this request.

    VPCId
        When creating a private hosted zone, either the VPC ID or VPC Name to associate with is
        required.  Exclusive with VPCName.  Ignored if passed for a non-private zone.

    VPCName
        When creating a private hosted zone, either the VPC ID or VPC Name to associate with is
        required.  Exclusive with VPCId.  Ignored if passed for a non-private zone.

    VPCRegion
        When creating a private hosted zone, the region of the associated VPC is required.  If not
        provided, an effort will be made to determine it from VPCId or VPCName, if possible.  If
        this fails, you'll need to provide an explicit value for this option.  Ignored if passed for
        a non-private zone.

    CallerReference
        A unique string that identifies the request and that allows create_hosted_zone() calls to be
        retried without the risk of executing the operation twice.  This is a required parameter
        when creating new Hosted Zones.  Maximum length of 128.

    Comment
        Any comments you want to include about the hosted zone.

    PrivateZone
        Boolean - Set to True if creating a private hosted zone.

    DelegationSetId
        If you want to associate a reusable delegation set with this hosted zone, the ID that Amazon
        Route 53 assigned to the reusable delegation set when you created it.  Note that XXX TODO
        create_delegation_set() is not yet implemented, so you'd need to manually create any
        delegation sets before utilizing this.

    region
        Region endpoint to connect to.

    key
        AWS key to bind with.

    keyid
        AWS keyid to bind with.

    profile
        Dict, or pillar key pointing to a dict, containing AWS region/key/keyid.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.create_hosted_zone example.org.
    """
    if not Name.endswith("."):
        raise SaltInvocationError(
            "Domain must be fully-qualified, complete with trailing period."
        )
    Name = _aws_encode(Name)
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    deets = find_hosted_zone(
        Name=Name,
        PrivateZone=PrivateZone,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )
    if deets:
        log.info(
            "Route 53 hosted zone %s already exists. You may want to pass "
            "e.g. 'PrivateZone=True' or similar...",
            Name,
        )
        return None
    args = {
        "Name": Name,
        "CallerReference": CallerReference,
        "HostedZoneConfig": {"Comment": Comment, "PrivateZone": PrivateZone},
    }
    args.update({"DelegationSetId": DelegationSetId}) if DelegationSetId else None
    if PrivateZone:
        if not _exactly_one((VPCName, VPCId)):
            raise SaltInvocationError(
                "Either VPCName or VPCId is required when creating a private zone."
            )
        vpcs = __salt__["boto_vpc.describe_vpcs"](
            vpc_id=VPCId,
            name=VPCName,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        ).get("vpcs", [])
        if VPCRegion and vpcs:
            vpcs = [v for v in vpcs if v["region"] == VPCRegion]
        if not vpcs:
            log.error(
                "Private zone requested but no VPC matching given criteria found."
            )
            return None
        if len(vpcs) > 1:
            log.error(
                "Private zone requested but multiple VPCs matching given "
                "criteria found: %s.",
                [v["id"] for v in vpcs],
            )
            return None
        vpc = vpcs[0]
        if VPCName:
            VPCId = vpc["id"]
        if not VPCRegion:
            VPCRegion = vpc["region"]
        args.update({"VPC": {"VPCId": VPCId, "VPCRegion": VPCRegion}})
    else:
        if any((VPCId, VPCName, VPCRegion)):
            log.info(
                "Options VPCId, VPCName, and VPCRegion are ignored when creating "
                "non-private zones."
            )
    tries = 10
    while tries:
        try:
            r = conn.create_hosted_zone(**args)
            r.pop("ResponseMetadata", None)
            if _wait_for_sync(r["ChangeInfo"]["Id"], conn):
                return [r]
            return []
        except ClientError as e:
            if tries and e.response.get("Error", {}).get("Code") == "Throttling":
                log.debug("Throttled by AWS API.")
                time.sleep(3)
                tries -= 1
                continue
            log.error("Failed to create hosted zone %s: %s", Name, e)
            return []
    return []


def update_hosted_zone_comment(
    Id=None,
    Name=None,
    Comment=None,
    PrivateZone=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Update the comment on an existing Route 53 hosted zone.

    Id
        The unique Zone Identifier for the Hosted Zone.

    Name
        The domain name associated with the Hosted Zone(s).

    Comment
        Any comments you want to include about the hosted zone.

    PrivateZone
        Boolean - Set to True if changing a private hosted zone.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.update_hosted_zone_comment Name=example.org. \
                Comment="This is an example comment for an example zone"
    """
    if not _exactly_one((Id, Name)):
        raise SaltInvocationError("Exactly one of either Id or Name is required.")
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if Name:
        args = {
            "Name": Name,
            "PrivateZone": PrivateZone,
            "region": region,
            "key": key,
            "keyid": keyid,
            "profile": profile,
        }
        zone = find_hosted_zone(**args)
        if not zone:
            log.error("Couldn't resolve domain name %s to a hosted zone ID.", Name)
            return []
        Id = zone[0]["HostedZone"]["Id"]
    tries = 10
    while tries:
        try:
            r = conn.update_hosted_zone_comment(Id=Id, Comment=Comment)
            r.pop("ResponseMetadata", None)
            return [r]
        except ClientError as e:
            if tries and e.response.get("Error", {}).get("Code") == "Throttling":
                log.debug("Throttled by AWS API.")
                time.sleep(3)
                tries -= 1
                continue
            log.error("Failed to update comment on hosted zone %s: %s", Name or Id, e)
    return []


def associate_vpc_with_hosted_zone(
    HostedZoneId=None,
    Name=None,
    VPCId=None,
    VPCName=None,
    VPCRegion=None,
    Comment=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Associates an Amazon VPC with a private hosted zone.

    To perform the association, the VPC and the private hosted zone must already exist. You can't
    convert a public hosted zone into a private hosted zone.  If you want to associate a VPC from
    one AWS account with a zone from a another, the AWS account owning the hosted zone must first
    submit a CreateVPCAssociationAuthorization (using create_vpc_association_authorization() or by
    other means, such as the AWS console).  With that done, the account owning the VPC can then call
    associate_vpc_with_hosted_zone() to create the association.

    Note that if both sides happen to be within the same account, associate_vpc_with_hosted_zone()
    is enough on its own, and there is no need for the CreateVPCAssociationAuthorization step.

    Also note that looking up hosted zones by name (e.g. using the Name parameter) only works
    within a single account - if you're associating a VPC to a zone in a different account, as
    outlined above, you unfortunately MUST use the HostedZoneId parameter exclusively.

    HostedZoneId
        The unique Zone Identifier for the Hosted Zone.

    Name
        The domain name associated with the Hosted Zone(s).

    VPCId
        When working with a private hosted zone, either the VPC ID or VPC Name to associate with is
        required.  Exclusive with VPCName.

    VPCName
        When working with a private hosted zone, either the VPC ID or VPC Name to associate with is
        required.  Exclusive with VPCId.

    VPCRegion
        When working with a private hosted zone, the region of the associated VPC is required.  If
        not provided, an effort will be made to determine it from VPCId or VPCName, if possible.  If
        this fails, you'll need to provide an explicit value for VPCRegion.

    Comment
        Any comments you want to include about the change being made.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.associate_vpc_with_hosted_zone \
                    Name=example.org. VPCName=myVPC \
                    VPCRegion=us-east-1 Comment="Whoo-hoo!  I added another VPC."

    """
    if not _exactly_one((HostedZoneId, Name)):
        raise SaltInvocationError(
            "Exactly one of either HostedZoneId or Name is required."
        )
    if not _exactly_one((VPCId, VPCName)):
        raise SaltInvocationError("Exactly one of either VPCId or VPCName is required.")
    if Name:
        # {'PrivateZone': True} because you can only associate VPCs with private hosted zones.
        args = {
            "Name": Name,
            "PrivateZone": True,
            "region": region,
            "key": key,
            "keyid": keyid,
            "profile": profile,
        }
        zone = find_hosted_zone(**args)
        if not zone:
            log.error(
                "Couldn't resolve domain name %s to a private hosted zone ID.", Name
            )
            return False
        HostedZoneId = zone[0]["HostedZone"]["Id"]
    vpcs = __salt__["boto_vpc.describe_vpcs"](
        vpc_id=VPCId, name=VPCName, region=region, key=key, keyid=keyid, profile=profile
    ).get("vpcs", [])
    if VPCRegion and vpcs:
        vpcs = [v for v in vpcs if v["region"] == VPCRegion]
    if not vpcs:
        log.error("No VPC matching the given criteria found.")
        return False
    if len(vpcs) > 1:
        log.error(
            "Multiple VPCs matching the given criteria found: %s.",
            ", ".join([v["id"] for v in vpcs]),
        )
        return False
    vpc = vpcs[0]
    if VPCName:
        VPCId = vpc["id"]
    if not VPCRegion:
        VPCRegion = vpc["region"]
    args = {
        "HostedZoneId": HostedZoneId,
        "VPC": {"VPCId": VPCId, "VPCRegion": VPCRegion},
    }
    args.update({"Comment": Comment}) if Comment is not None else None

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    tries = 10
    while tries:
        try:
            r = conn.associate_vpc_with_hosted_zone(**args)
            return _wait_for_sync(r["ChangeInfo"]["Id"], conn)
        except ClientError as e:
            if e.response.get("Error", {}).get("Code") == "ConflictingDomainExists":
                log.debug("VPC Association already exists.")
                # return True since the current state is the desired one
                return True
            if tries and e.response.get("Error", {}).get("Code") == "Throttling":
                log.debug("Throttled by AWS API.")
                time.sleep(3)
                tries -= 1
                continue
            log.error(
                "Failed to associate VPC %s with hosted zone %s: %s",
                VPCName or VPCId,
                Name or HostedZoneId,
                e,
            )
    return False


def disassociate_vpc_from_hosted_zone(
    HostedZoneId=None,
    Name=None,
    VPCId=None,
    VPCName=None,
    VPCRegion=None,
    Comment=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Disassociates an Amazon VPC from a private hosted zone.

    You can't disassociate the last VPC from a private hosted zone.  You also can't convert a
    private hosted zone into a public hosted zone.

    Note that looking up hosted zones by name (e.g. using the Name parameter) only works XXX FACTCHECK
    within a single AWS account - if you're disassociating a VPC in one account from a hosted zone
    in a different account you unfortunately MUST use the HostedZoneId parameter exclusively. XXX FIXME DOCU

    HostedZoneId
        The unique Zone Identifier for the Hosted Zone.

    Name
        The domain name associated with the Hosted Zone(s).

    VPCId
        When working with a private hosted zone, either the VPC ID or VPC Name to associate with is
        required.  Exclusive with VPCName.

    VPCName
        When working with a private hosted zone, either the VPC ID or VPC Name to associate with is
        required.  Exclusive with VPCId.

    VPCRegion
        When working with a private hosted zone, the region of the associated VPC is required.  If
        not provided, an effort will be made to determine it from VPCId or VPCName, if possible.  If
        this fails, you'll need to provide an explicit value for VPCRegion.

    Comment
        Any comments you want to include about the change being made.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.disassociate_vpc_from_hosted_zone \
                    Name=example.org. VPCName=myVPC \
                    VPCRegion=us-east-1 Comment="Whoops!  Don't wanna talk to this-here zone no more."

    """
    if not _exactly_one((HostedZoneId, Name)):
        raise SaltInvocationError(
            "Exactly one of either HostedZoneId or Name is required."
        )
    if not _exactly_one((VPCId, VPCName)):
        raise SaltInvocationError("Exactly one of either VPCId or VPCName is required.")
    if Name:
        # {'PrivateZone': True} because you can only associate VPCs with private hosted zones.
        args = {
            "Name": Name,
            "PrivateZone": True,
            "region": region,
            "key": key,
            "keyid": keyid,
            "profile": profile,
        }
        zone = find_hosted_zone(**args)
        if not zone:
            log.error(
                "Couldn't resolve domain name %s to a private hosted zone ID.", Name
            )
            return False
        HostedZoneId = zone[0]["HostedZone"]["Id"]
    vpcs = __salt__["boto_vpc.describe_vpcs"](
        vpc_id=VPCId, name=VPCName, region=region, key=key, keyid=keyid, profile=profile
    ).get("vpcs", [])
    if VPCRegion and vpcs:
        vpcs = [v for v in vpcs if v["region"] == VPCRegion]
    if not vpcs:
        log.error("No VPC matching the given criteria found.")
        return False
    if len(vpcs) > 1:
        log.error(
            "Multiple VPCs matching the given criteria found: %s.",
            ", ".join([v["id"] for v in vpcs]),
        )
        return False
    vpc = vpcs[0]
    if VPCName:
        VPCId = vpc["id"]
    if not VPCRegion:
        VPCRegion = vpc["region"]
    args = {
        "HostedZoneId": HostedZoneId,
        "VPC": {"VPCId": VPCId, "VPCRegion": VPCRegion},
    }
    args.update({"Comment": Comment}) if Comment is not None else None

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    tries = 10
    while tries:
        try:
            r = conn.disassociate_vpc_from_hosted_zone(**args)
            return _wait_for_sync(r["ChangeInfo"]["Id"], conn)
        except ClientError as e:
            if e.response.get("Error", {}).get("Code") == "VPCAssociationNotFound":
                log.debug("No VPC Association exists.")
                # return True since the current state is the desired one
                return True
            if tries and e.response.get("Error", {}).get("Code") == "Throttling":
                log.debug("Throttled by AWS API.")
                time.sleep(3)
                tries -= 1
                continue
            log.error(
                "Failed to associate VPC %s with hosted zone %s: %s",
                VPCName or VPCId,
                Name or HostedZoneId,
                e,
            )
    return False


# def create_vpc_association_authorization(*args, **kwargs):
#    '''
#    unimplemented
#    '''
#    pass


# def delete_vpc_association_authorization(*args, **kwargs):
#    '''
#    unimplemented
#    '''
#    pass


# def list_vpc_association_authorizations(*args, **kwargs):
#    '''
#    unimplemented
#    '''
#    pass


def delete_hosted_zone(Id, region=None, key=None, keyid=None, profile=None):
    """
    Delete a Route53 hosted zone.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.delete_hosted_zone Z1234567890
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        r = conn.delete_hosted_zone(Id=Id)
        return _wait_for_sync(r["ChangeInfo"]["Id"], conn)
    except ClientError as e:
        log.error("Failed to delete hosted zone %s: %s", Id, e)
    return False


def delete_hosted_zone_by_domain(
    Name, PrivateZone=None, region=None, key=None, keyid=None, profile=None
):
    """
    Delete a Route53 hosted zone by domain name, and PrivateZone status if provided.

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.delete_hosted_zone_by_domain example.org.
    """
    args = {
        "Name": Name,
        "PrivateZone": PrivateZone,
        "region": region,
        "key": key,
        "keyid": keyid,
        "profile": profile,
    }
    # Be extra pedantic in the service of safety - if public/private is not provided and the domain
    # name resolves to both, fail and require them to declare it explicitly.
    zone = find_hosted_zone(**args)
    if not zone:
        log.error("Couldn't resolve domain name %s to a hosted zone ID.", Name)
        return False
    Id = zone[0]["HostedZone"]["Id"]
    return delete_hosted_zone(
        Id=Id, region=region, key=key, keyid=keyid, profile=profile
    )


def _aws_encode(x):
    """
    An implementation of the encoding required to support AWS's domain name
    rules defined here__:

    .. __: http://docs.aws.amazon.com/Route53/latest/DeveloperGuide/DomainNameFormat.html

    While AWS's documentation specifies individual ASCII characters which need
    to be encoded, we instead just try to force the string to one of
    escaped unicode or idna depending on whether there are non-ASCII characters
    present.

    This means that we support things like ドメイン.テスト as a domain name string.

    More information about IDNA encoding in python is found here__:

    .. __: https://pypi.org/project/idna

    """
    ret = None
    try:
        x.encode("ascii")
        ret = re.sub(rb"\\x([a-f0-8]{2})", _hexReplace, x.encode("unicode_escape"))
    except UnicodeEncodeError:
        ret = x.encode("idna")
    except Exception as e:  # pylint: disable=broad-except
        log.error(
            "Couldn't encode %s using either 'unicode_escape' or 'idna' codecs", x
        )
        raise CommandExecutionError(e)
    log.debug("AWS-encoded result for %s: %s", x, ret)
    return ret.decode("utf-8")


def _aws_encode_changebatch(o):
    """
    helper method to process a change batch & encode the bits which need encoding.
    """
    change_idx = 0
    while change_idx < len(o["Changes"]):
        o["Changes"][change_idx]["ResourceRecordSet"]["Name"] = _aws_encode(
            o["Changes"][change_idx]["ResourceRecordSet"]["Name"]
        )
        if "ResourceRecords" in o["Changes"][change_idx]["ResourceRecordSet"]:
            rr_idx = 0
            while rr_idx < len(
                o["Changes"][change_idx]["ResourceRecordSet"]["ResourceRecords"]
            ):
                o["Changes"][change_idx]["ResourceRecordSet"]["ResourceRecords"][
                    rr_idx
                ]["Value"] = _aws_encode(
                    o["Changes"][change_idx]["ResourceRecordSet"]["ResourceRecords"][
                        rr_idx
                    ]["Value"]
                )
                rr_idx += 1
        if "AliasTarget" in o["Changes"][change_idx]["ResourceRecordSet"]:
            o["Changes"][change_idx]["ResourceRecordSet"]["AliasTarget"]["DNSName"] = (
                _aws_encode(
                    o["Changes"][change_idx]["ResourceRecordSet"]["AliasTarget"][
                        "DNSName"
                    ]
                )
            )
        change_idx += 1
    return o


def _aws_decode(x):
    """
    An implementation of the decoding required to support AWS's domain name
    rules defined here__:

    .. __: http://docs.aws.amazon.com/Route53/latest/DeveloperGuide/DomainNameFormat.html

    The important part is this:

        If the domain name includes any characters other than a to z, 0 to 9, - (hyphen),
        or _ (underscore), Route 53 API actions return the characters as escape codes.
        This is true whether you specify the characters as characters or as escape
        codes when you create the entity.
        The Route 53 console displays the characters as characters, not as escape codes."

        For a list of ASCII characters the corresponding octal codes, do an internet search on "ascii table".

    We look for the existence of any escape codes which give us a clue that
    we're received an escaped unicode string; or we assume it's idna encoded
    and then decode as necessary.
    """
    if "\\" in x:
        return x.decode("unicode_escape")

    if isinstance(x, bytes):
        return x.decode("idna")

    return x


def _hexReplace(x):
    """
    Converts a hex code to a base 16 int then the octal of it, minus the leading
    zero.

    This is necessary because x.encode('unicode_escape') automatically assumes
    you want a hex string, which AWS will accept but doesn't result in what
    you really want unless it's an octal escape sequence
    """
    c = int(x.group(1), 16)
    return "\\" + str(oct(c))[1:]


def get_resource_records(
    HostedZoneId=None,
    Name=None,
    StartRecordName=None,
    StartRecordType=None,
    PrivateZone=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Get all resource records from a given zone matching the provided StartRecordName (if given) or all
    records in the zone (if not), optionally filtered by a specific StartRecordType.  This will return
    any and all RRs matching, regardless of their special AWS flavors (weighted, geolocation, alias,
    etc.) so your code should be prepared for potentially large numbers of records back from this
    function - for example, if you've created a complex geolocation mapping with lots of entries all
    over the world providing the same server name to many different regional clients.

    If you want EXACTLY ONE record to operate on, you'll need to implement any logic required to
    pick the specific RR you care about from those returned.

    Note that if you pass in Name without providing a value for PrivateZone (either True or
    False), CommandExecutionError can be raised in the case of both public and private zones
    matching the domain. XXX FIXME DOCU

    CLI Example:

    .. code-block:: bash

        salt myminion boto3_route53.get_records test.example.org example.org A
    """
    if not _exactly_one((HostedZoneId, Name)):
        raise SaltInvocationError(
            "Exactly one of either HostedZoneId or Name must be provided."
        )
    if Name:
        args = {
            "Name": Name,
            "region": region,
            "key": key,
            "keyid": keyid,
            "profile": profile,
        }
        args.update({"PrivateZone": PrivateZone}) if PrivateZone is not None else None
        zone = find_hosted_zone(**args)
        if not zone:
            log.error("Couldn't resolve domain name %s to a hosted zone ID.", Name)
            return []
        HostedZoneId = zone[0]["HostedZone"]["Id"]

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    ret = []
    next_rr_name = StartRecordName
    next_rr_type = StartRecordType
    next_rr_id = None
    done = False
    while True:
        if done:
            return ret
        args = {"HostedZoneId": HostedZoneId}
        (
            args.update({"StartRecordName": _aws_encode(next_rr_name)})
            if next_rr_name
            else None
        )
        # Grrr, can't specify type unless name is set...  We'll do this via filtering later instead
        (
            args.update({"StartRecordType": next_rr_type})
            if next_rr_name and next_rr_type
            else None
        )
        args.update({"StartRecordIdentifier": next_rr_id}) if next_rr_id else None
        try:
            r = conn.list_resource_record_sets(**args)
            rrs = r["ResourceRecordSets"]
            next_rr_name = r.get("NextRecordName")
            next_rr_type = r.get("NextRecordType")
            next_rr_id = r.get("NextRecordIdentifier")
            for rr in rrs:
                rr["Name"] = _aws_decode(rr["Name"])
                # now iterate over the ResourceRecords and replace any encoded
                # value strings with the decoded versions
                if "ResourceRecords" in rr:
                    x = 0
                    while x < len(rr["ResourceRecords"]):
                        if "Value" in rr["ResourceRecords"][x]:
                            rr["ResourceRecords"][x]["Value"] = _aws_decode(
                                rr["ResourceRecords"][x]["Value"]
                            )
                        x += 1
                # or if we are an AliasTarget then decode the DNSName
                if "AliasTarget" in rr:
                    rr["AliasTarget"]["DNSName"] = _aws_decode(
                        rr["AliasTarget"]["DNSName"]
                    )
                if StartRecordName and rr["Name"] != StartRecordName:
                    done = True
                    break
                if StartRecordType and rr["Type"] != StartRecordType:
                    if StartRecordName:
                        done = True
                        break
                    else:
                        # We're filtering by type alone, and there might be more later, so...
                        continue
                ret += [rr]
            if not next_rr_name:
                done = True
        except ClientError as e:
            # Try forever on a simple thing like this...
            if e.response.get("Error", {}).get("Code") == "Throttling":
                log.debug("Throttled by AWS API.")
                time.sleep(3)
                continue
            raise


def change_resource_record_sets(
    HostedZoneId=None,
    Name=None,
    PrivateZone=None,
    ChangeBatch=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    See the `AWS Route53 API docs`__ as well as the `Boto3 documentation`__ for all the details...

    .. __: https://docs.aws.amazon.com/Route53/latest/APIReference/API_ChangeResourceRecordSets.html
    .. __: http://boto3.readthedocs.io/en/latest/reference/services/route53.html#Route53.Client.change_resource_record_sets

    The syntax for a ChangeBatch parameter is as follows, but note that the permutations of allowed
    parameters and combinations thereof are quite varied, so perusal of the above linked docs is
    highly recommended for any non-trival configurations.

    .. code-block:: text

        {
            "Comment": "string",
            "Changes": [
                {
                    "Action": "CREATE"|"DELETE"|"UPSERT",
                    "ResourceRecordSet": {
                        "Name": "string",
                        "Type": "SOA"|"A"|"TXT"|"NS"|"CNAME"|"MX"|"NAPTR"|"PTR"|"SRV"|"SPF"|"AAAA",
                        "SetIdentifier": "string",
                        "Weight": 123,
                        "Region": "us-east-1"|"us-east-2"|"us-west-1"|"us-west-2"|"ca-central-1"|"eu-west-1"|"eu-west-2"|"eu-central-1"|"ap-southeast-1"|"ap-southeast-2"|"ap-northeast-1"|"ap-northeast-2"|"sa-east-1"|"cn-north-1"|"ap-south-1",
                        "GeoLocation": {
                            "ContinentCode": "string",
                            "CountryCode": "string",
                            "SubdivisionCode": "string"
                        },
                        "Failover": "PRIMARY"|"SECONDARY",
                        "TTL": 123,
                        "ResourceRecords": [
                            {
                                "Value": "string"
                            },
                        ],
                        "AliasTarget": {
                            "HostedZoneId": "string",
                            "DNSName": "string",
                            "EvaluateTargetHealth": True|False
                        },
                        "HealthCheckId": "string",
                        "TrafficPolicyInstanceId": "string"
                    }
                },
            ]
        }

    CLI Example:

    .. code-block:: bash

        foo='{
               "Name": "my-cname.example.org.",
               "TTL": 600,
               "Type": "CNAME",
               "ResourceRecords": [
                 {
                   "Value": "my-host.example.org"
                 }
               ]
             }'
        foo=`echo $foo`  # Remove newlines
        salt myminion boto3_route53.change_resource_record_sets DomainName=example.org. \
                keyid=A1234567890ABCDEF123 key=xblahblahblah \
                ChangeBatch="{'Changes': [{'Action': 'UPSERT', 'ResourceRecordSet': $foo}]}"
    """
    if not _exactly_one((HostedZoneId, Name)):
        raise SaltInvocationError(
            "Exactly one of either HostZoneId or Name must be provided."
        )
    if Name:
        args = {
            "Name": Name,
            "region": region,
            "key": key,
            "keyid": keyid,
            "profile": profile,
        }
        args.update({"PrivateZone": PrivateZone}) if PrivateZone is not None else None
        zone = find_hosted_zone(**args)
        if not zone:
            log.error("Couldn't resolve domain name %s to a hosted zone ID.", Name)
            return []
        HostedZoneId = zone[0]["HostedZone"]["Id"]

    args = {
        "HostedZoneId": HostedZoneId,
        "ChangeBatch": _aws_encode_changebatch(ChangeBatch),
    }

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    tries = 20  # A bit more headroom
    while tries:
        try:
            r = conn.change_resource_record_sets(**args)
            return _wait_for_sync(
                r["ChangeInfo"]["Id"], conn, 30
            )  # And a little extra time here
        except ClientError as e:
            if tries and e.response.get("Error", {}).get("Code") == "Throttling":
                log.debug("Throttled by AWS API.")
                time.sleep(3)
                tries -= 1
                continue
            log.error(
                "Failed to apply requested changes to the hosted zone %s: %s",
                (Name or HostedZoneId),
                str(e),
            )
            raise e
    return False



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.linux_service
"""

import os

import pytest

import salt.modules.linux_service as service
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {service: {}}


def test_start():
    """
    Test to start the specified service
    """
    with patch.object(os.path, "join", return_value="A"):
        with patch.object(service, "run", MagicMock(return_value=True)):
            assert service.start("name")


def test_stop():
    """
    Test to stop the specified service
    """
    with patch.object(os.path, "join", return_value="A"):
        with patch.object(service, "run", MagicMock(return_value=True)):
            assert service.stop("name")


def test_restart():
    """
    Test to restart the specified service
    """
    with patch.object(os.path, "join", return_value="A"):
        with patch.object(service, "run", MagicMock(return_value=True)):
            assert service.restart("name")


def test_status():
    """
    Test to return the status for a service, returns the PID or an empty
    string if the service is running or not, pass a signature to use to
    find the service via ps
    """
    with patch.dict(service.__salt__, {"status.pid": MagicMock(return_value=True)}):
        assert service.status("name")


def test_reload_():
    """
    Test to restart the specified service
    """
    with patch.object(os.path, "join", return_value="A"):
        with patch.object(service, "run", MagicMock(return_value=True)):
            assert service.reload_("name")


def test_run():
    """
    Test to run the specified service
    """
    with patch.object(os.path, "join", return_value="A"):
        with patch.object(service, "run", MagicMock(return_value=True)):
            assert service.run("name", "action")


def test_available():
    """
    Test to returns ``True`` if the specified service is available,
    otherwise returns ``False``.
    """
    with patch.object(service, "get_all", return_value=["name", "A"]):
        assert service.available("name")


def test_missing():
    """
    Test to inverse of service.available.
    """
    with patch.object(service, "get_all", return_value=["name1", "A"]):
        assert service.missing("name")


def test_get_all():
    """
    Test to return a list of all available services
    """
    with patch.object(os.path, "isdir", side_effect=[False, True]):

        assert service.get_all() == []

        with patch.object(os, "listdir", return_value=["A", "B"]):
            assert service.get_all() == ["A", "B"]



"""
Integration tests for the lxd states
"""

import pytest

import salt.modules.lxd
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin


@pytest.mark.skipif(salt.modules.lxd.HAS_PYLXD is False, reason="pylxd not installed")
@pytest.mark.skip_if_binaries_missing("lxd", reason="LXD not installed")
@pytest.mark.skip_if_binaries_missing("lxc", reason="LXC not installed")
class LxdImageTestCase(ModuleCase, SaltReturnAssertsMixin):
    def test_02__pull_image(self):
        ret = self.run_state(
            "lxd_image.present",
            name="images:centos/7",
            source={
                "name": "centos/7",
                "type": "simplestreams",
                "server": "https://images.linuxcontainers.org",
            },
        )
        name = "lxd_image_|-images:centos/7_|-images:centos/7_|-present"
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]["changes"]["aliases"] == ['Added alias "images:centos/7"']

    def test_03__delete_image(self):
        ret = self.run_state(
            "lxd_image.absent",
            name="images:centos/7",
        )
        name = "lxd_image_|-images:centos/7_|-images:centos/7_|-absent"
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert (
            ret[name]["changes"]["removed"]
            == 'Image "images:centos/7" has been deleted.'
        )



"""
This module contains all of the routines needed to set up a master server, this
involves preparing the three listeners and the workers needed by the master.
"""

import fnmatch
import logging
import os
import re
import stat
import time

import salt.acl
import salt.auth
import salt.cache
import salt.client
import salt.crypt
import salt.exceptions
import salt.fileserver
import salt.key
import salt.minion
import salt.payload
import salt.pillar
import salt.runner
import salt.state
import salt.utils.args
import salt.utils.atomicfile
import salt.utils.dictupdate
import salt.utils.event
import salt.utils.files
import salt.utils.gitfs
import salt.utils.gzip_util
import salt.utils.jid
import salt.utils.job
import salt.utils.mine
import salt.utils.minions
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.user
import salt.utils.verify
import salt.wheel
from salt.defaults import DEFAULT_TARGET_DELIM
from salt.pillar import git_pillar

try:
    import pwd

    HAS_PWD = True
except ImportError:
    # pwd is not available on windows
    HAS_PWD = False

log = logging.getLogger(__name__)

# Things to do in lower layers:
# only accept valid minion ids


def init_git_pillar(opts):
    """
    Clear out the ext pillar caches, used when the master starts
    """
    ret = []
    for opts_dict in [x for x in opts.get("ext_pillar", [])]:
        if "git" in opts_dict:
            try:
                pillar = salt.utils.gitfs.GitPillar(
                    opts,
                    opts_dict["git"],
                    per_remote_overrides=git_pillar.PER_REMOTE_OVERRIDES,
                    per_remote_only=git_pillar.PER_REMOTE_ONLY,
                    global_only=git_pillar.GLOBAL_ONLY,
                )
                ret.append(pillar)
            except salt.exceptions.FileserverConfigError:
                if opts.get("git_pillar_verify_config", True):
                    raise
                else:
                    log.critical("Could not initialize git_pillar")
    return ret


def clean_fsbackend(opts):
    """
    Clean out the old fileserver backends
    """
    # Clear remote fileserver backend caches so they get recreated
    for backend in ("git", "hg", "svn"):
        if backend in opts["fileserver_backend"]:
            env_cache = os.path.join(opts["cachedir"], f"{backend}fs", "envs.p")
            if os.path.isfile(env_cache):
                log.debug("Clearing %sfs env cache", backend)
                try:
                    os.remove(env_cache)
                except OSError as exc:
                    log.critical(
                        "Unable to clear env cache file %s: %s", env_cache, exc
                    )

            file_lists_dir = os.path.join(
                opts["cachedir"], "file_lists", f"{backend}fs"
            )
            try:
                file_lists_caches = os.listdir(file_lists_dir)
            except OSError:
                continue
            for file_lists_cache in fnmatch.filter(file_lists_caches, "*.p"):
                cache_file = os.path.join(file_lists_dir, file_lists_cache)
                try:
                    os.remove(cache_file)
                except OSError as exc:
                    log.critical(
                        "Unable to file_lists cache file %s: %s", cache_file, exc
                    )


def clean_expired_tokens(opts):
    """
    Clean expired tokens from the master
    """
    loadauth = salt.auth.LoadAuth(opts)
    for tok in loadauth.list_tokens():
        token_data = loadauth.get_tok(tok)
        if "expire" not in token_data or token_data.get("expire", 0) < time.time():
            loadauth.rm_token(tok)


def clean_pub_auth(opts):
    try:
        auth_cache = os.path.join(opts["cachedir"], "publish_auth")
        if not os.path.exists(auth_cache):
            return
        else:
            for dirpath, dirnames, filenames in salt.utils.path.os_walk(auth_cache):
                for auth_file in filenames:
                    auth_file_path = os.path.join(dirpath, auth_file)
                    if not os.path.isfile(auth_file_path):
                        continue
                    if time.time() - os.path.getmtime(auth_file_path) > (
                        salt.utils.job.get_keep_jobs_seconds(opts)
                    ):
                        os.remove(auth_file_path)
    except OSError:
        log.error("Unable to delete pub auth file")


def clean_old_jobs(opts):
    """
    Clean out the old jobs from the job cache
    """
    # TODO: better way to not require creating the masterminion every time?
    mminion = salt.minion.MasterMinion(
        opts,
        states=False,
        rend=False,
    )
    # If the master job cache has a clean_old_jobs, call it
    fstr = "{}.clean_old_jobs".format(opts["master_job_cache"])
    if fstr in mminion.returners:
        mminion.returners[fstr]()


def mk_key(opts, user):
    if HAS_PWD:
        uid = None
        try:
            uid = pwd.getpwnam(user).pw_uid
        except KeyError:
            # User doesn't exist in the system
            if opts["client_acl_verify"]:
                return None
    if salt.utils.platform.is_windows():
        # The username may contain '\' if it is in Windows
        # 'DOMAIN\username' format. Fix this for the keyfile path.
        keyfile = os.path.join(
            opts["cachedir"], ".{}_key".format(user.replace("\\", "_"))
        )
    else:
        keyfile = os.path.join(opts["cachedir"], f".{user}_key")

    if os.path.exists(keyfile):
        log.debug("Removing stale keyfile: %s", keyfile)
        if salt.utils.platform.is_windows() and not os.access(keyfile, os.W_OK):
            # Cannot delete read-only files on Windows.
            os.chmod(keyfile, stat.S_IRUSR | stat.S_IWUSR)
        os.unlink(keyfile)

    key = salt.crypt.Crypticle.generate_key_string()
    with salt.utils.files.set_umask(0o277):
        with salt.utils.files.fopen(keyfile, "w+") as fp_:
            fp_.write(salt.utils.stringutils.to_str(key))
    # 600 octal: Read and write access to the owner only.
    # Write access is necessary since on subsequent runs, if the file
    # exists, it needs to be written to again. Windows enforces this.
    os.chmod(keyfile, 0o600)
    if HAS_PWD and uid is not None:
        try:
            os.chown(keyfile, uid, -1)
        except OSError:
            # The master is not being run as root and can therefore not
            # chown the key file
            pass
    return key


def access_keys(opts):
    """
    A key needs to be placed in the filesystem with permissions 0400 so
    clients are required to run as root.
    """
    # TODO: Need a way to get all available users for systems not supported by pwd module.
    #       For now users pattern matching will not work for publisher_acl.
    keys = {}
    publisher_acl = opts["publisher_acl"]
    acl_users = set(publisher_acl.keys())
    if opts.get("user"):
        acl_users.add(opts["user"])
    acl_users.add(salt.utils.user.get_user())
    for user in acl_users:
        log.info("Preparing the %s key for local communication", user)

        keyfile = os.path.join(opts["cachedir"], f".{user}_key")
        if os.path.exists(keyfile):
            with salt.utils.files.fopen(keyfile, "r") as fp:
                key = salt.utils.stringutils.to_unicode(fp.read())
        else:
            key = mk_key(opts, user)
        if key is not None:
            keys[user] = key

    # Check other users matching ACL patterns
    if opts["client_acl_verify"] and HAS_PWD:
        log.profile("Beginning pwd.getpwall() call in masterapi access_keys function")
        for user in pwd.getpwall():
            user = user.pw_name
            if user not in keys and salt.utils.stringutils.check_whitelist_blacklist(
                user, whitelist=acl_users
            ):
                if os.path.exists(keyfile):
                    with salt.utils.files.fopen(keyfile, "r") as fp:
                        keys[user] = salt.utils.stringutils.to_unicode(fp.read())
                else:
                    keys[user] = mk_key(opts, user)
        log.profile("End pwd.getpwall() call in masterapi access_keys function")

    return keys


def fileserver_update(fileserver):
    """
    Update the fileserver backends, requires that a salt.fileserver.Fileserver
    object be passed in
    """
    try:
        if not fileserver.servers:
            log.error(
                "No fileservers loaded, the master will not be able to "
                "serve files to minions"
            )
            raise salt.exceptions.SaltMasterError("No fileserver backends available")
        fileserver.update()
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "Exception %s occurred in file server update",
            exc,
            exc_info_on_loglevel=logging.DEBUG,
        )


class AutoKey:
    """
    Implement the methods to run auto key acceptance and rejection
    """

    def __init__(self, opts):
        self.signing_files = {}
        self.opts = opts

    def check_permissions(self, filename):
        """
        Check if the specified filename has correct permissions
        """
        if salt.utils.platform.is_windows():
            return True

        # After we've ascertained we're not on windows
        groups = salt.utils.user.get_gid_list(self.opts["user"], include_default=False)
        fmode = os.stat(filename)

        if stat.S_IWOTH & fmode.st_mode:
            # don't allow others to write to the file
            return False

        if stat.S_IWGRP & fmode.st_mode:
            # if the group has write access only allow with permissive_pki_access
            if not self.opts.get("permissive_pki_access", False):
                return False
            elif os.getuid() == 0 and fmode.st_gid not in groups:
                # if salt is root it has to be in the group that has write access
                # this gives the group 'permission' to have write access
                return False

        return True

    def check_signing_file(self, keyid, signing_file):
        """
        Check a keyid for membership in a signing file
        """
        if not signing_file or not os.path.exists(signing_file):
            return False

        if not self.check_permissions(signing_file):
            log.warning("Wrong permissions for %s, ignoring content", signing_file)
            return False

        mtime = os.path.getmtime(signing_file)
        if self.signing_files.get(signing_file, {}).get("mtime") != mtime:
            self.signing_files.setdefault(signing_file, {})["mtime"] = mtime
            with salt.utils.files.fopen(signing_file, "r") as fp_:
                self.signing_files[signing_file]["data"] = [
                    entry
                    for entry in [line.strip() for line in fp_]
                    if not entry.strip().startswith("#")
                ]
        return any(
            salt.utils.stringutils.expr_match(keyid, line)
            for line in self.signing_files[signing_file].get("data", [])
        )

    def check_autosign_dir(self, keyid):
        """
        Check a keyid for membership in a autosign directory.
        """
        if self.opts["cluster_id"]:
            pki_dir = self.opts["cluster_pki_dir"]
        else:
            pki_dir = self.opts["pki_dir"]
        autosign_dir = os.path.join(pki_dir, "minions_autosign")

        # cleanup expired files
        expire_minutes = self.opts.get("autosign_timeout", 120)
        if expire_minutes > 0:
            min_time = time.time() - (60 * int(expire_minutes))
            for root, dirs, filenames in salt.utils.path.os_walk(autosign_dir):
                for f in filenames:
                    stub_file = os.path.join(autosign_dir, f)
                    mtime = os.path.getmtime(stub_file)
                    if mtime < min_time:
                        log.warning("Autosign keyid expired %s", stub_file)
                        os.remove(stub_file)

        stub_file = os.path.join(autosign_dir, keyid)
        if not os.path.exists(stub_file):
            return False
        os.remove(stub_file)
        return True

    def check_autosign_grains(self, autosign_grains):
        """
        Check for matching grains in the autosign_grains_dir.
        """
        if not autosign_grains or "autosign_grains_dir" not in self.opts:
            return False

        autosign_grains_dir = self.opts["autosign_grains_dir"]
        for root, dirs, filenames in os.walk(autosign_grains_dir):
            for grain in filenames:
                if grain in autosign_grains:
                    grain_file = os.path.join(autosign_grains_dir, grain)

                    if not self.check_permissions(grain_file):
                        log.warning(
                            "Wrong permissions for %s, ignoring content", grain_file
                        )
                        continue

                    with salt.utils.files.fopen(grain_file, "r") as f:
                        for line in f:
                            line = salt.utils.stringutils.to_unicode(line).strip()
                            if line.startswith("#"):
                                continue
                            if str(autosign_grains[grain]) == line:
                                return True
        return False

    def check_autoreject(self, keyid):
        """
        Checks if the specified keyid should automatically be rejected.
        """
        return self.check_signing_file(keyid, self.opts.get("autoreject_file", None))

    def check_autosign(self, keyid, autosign_grains=None):
        """
        Checks if the specified keyid should automatically be signed.
        """
        if self.opts["auto_accept"]:
            return True
        if self.check_signing_file(keyid, self.opts.get("autosign_file", None)):
            return True
        if self.check_autosign_dir(keyid):
            return True
        if self.check_autosign_grains(autosign_grains):
            return True
        return False


class RemoteFuncs:
    """
    Funcitons made available to minions, this class includes the raw routines
    post validation that make up the minion access to the master
    """

    def __init__(self, opts):
        self.opts = opts
        self.event = salt.utils.event.get_event(
            "master",
            self.opts["sock_dir"],
            opts=self.opts,
            listen=False,
        )
        self.ckminions = salt.utils.minions.CkMinions(opts)
        # Create the tops dict for loading external top data
        self.tops = salt.loader.tops(self.opts)
        # Make a client
        self.local = salt.client.get_local_client(mopts=self.opts)
        # Create the master minion to access the external job cache
        self.mminion = salt.minion.MasterMinion(self.opts, states=False, rend=False)
        self.__setup_fileserver()
        self.cache = salt.cache.factory(opts)

    def __setup_fileserver(self):
        """
        Set the local file objects from the file server interface
        """
        fs_ = salt.fileserver.Fileserver(self.opts)
        self._serve_file = fs_.serve_file
        self._file_find = fs_._find_file
        self._file_hash = fs_.file_hash
        self._file_list = fs_.file_list
        self._file_list_emptydirs = fs_.file_list_emptydirs
        self._dir_list = fs_.dir_list
        self._symlink_list = fs_.symlink_list
        self._file_envs = fs_.envs

    def __verify_minion_publish(self, load):
        """
        Verify that the passed information authorized a minion to execute
        """
        # Verify that the load is valid
        if "peer" not in self.opts:
            return False
        if not isinstance(self.opts["peer"], dict):
            return False
        if any(key not in load for key in ("fun", "arg", "tgt", "ret", "id")):
            return False
        # If the command will make a recursive publish don't run
        if re.match("publish.*", load["fun"]):
            return False
        # Check the permissions for this minion
        perms = []
        for match in self.opts["peer"]:
            if re.match(match, load["id"]):
                # This is the list of funcs/modules!
                if isinstance(self.opts["peer"][match], list):
                    perms.extend(self.opts["peer"][match])
        if "," in load["fun"]:
            # 'arg': [['cat', '/proc/cpuinfo'], [], ['foo']]
            load["fun"] = load["fun"].split(",")
            arg_ = []
            for arg in load["arg"]:
                arg_.append(arg.split())
            load["arg"] = arg_
        return self.ckminions.auth_check(
            perms,
            load["fun"],
            load["arg"],
            load["tgt"],
            load.get("tgt_type", "glob"),
            publish_validate=True,
        )

    def _master_opts(self, load):
        """
        Return the master options to the minion
        """
        mopts = {}
        file_roots = {}
        envs = self._file_envs()
        for saltenv in envs:
            if saltenv not in file_roots:
                file_roots[saltenv] = []
        mopts["file_roots"] = file_roots
        mopts["top_file_merging_strategy"] = self.opts["top_file_merging_strategy"]
        mopts["env_order"] = self.opts["env_order"]
        mopts["default_top"] = self.opts["default_top"]
        if load.get("env_only"):
            return mopts
        mopts["renderer"] = self.opts["renderer"]
        mopts["failhard"] = self.opts["failhard"]
        mopts["state_top"] = self.opts["state_top"]
        mopts["state_top_saltenv"] = self.opts["state_top_saltenv"]
        mopts["nodegroups"] = self.opts["nodegroups"]
        mopts["state_auto_order"] = self.opts["state_auto_order"]
        mopts["state_events"] = self.opts["state_events"]
        mopts["state_aggregate"] = self.opts["state_aggregate"]
        mopts["jinja_env"] = self.opts["jinja_env"]
        mopts["jinja_sls_env"] = self.opts["jinja_sls_env"]
        mopts["jinja_lstrip_blocks"] = self.opts["jinja_lstrip_blocks"]
        mopts["jinja_trim_blocks"] = self.opts["jinja_trim_blocks"]
        return mopts

    def _master_tops(self, load, skip_verify=False):
        """
        Return the results from master_tops if configured
        """
        if not skip_verify:
            if "id" not in load:
                log.error("Received call for external nodes without an id")
                return {}
            if not salt.utils.verify.valid_id(self.opts, load["id"]):
                return {}
        # Evaluate all configured master_tops interfaces

        opts = {}
        grains = {}
        ret = {}

        if "opts" in load:
            opts = load["opts"]
            if "grains" in load["opts"]:
                grains = load["opts"]["grains"]
        for fun in self.tops:
            if fun not in self.opts.get("master_tops", {}):
                continue
            try:
                ret = salt.utils.dictupdate.merge(
                    ret, self.tops[fun](opts=opts, grains=grains), merge_lists=True
                )
            except Exception as exc:  # pylint: disable=broad-except
                # If anything happens in the top generation, log it and move on
                log.error(
                    "Top function %s failed with error %s for minion %s",
                    fun,
                    exc,
                    load["id"],
                )
        return ret

    def _mine_get(self, load, skip_verify=False):
        """
        Gathers the data from the specified minions' mine
        """
        if not skip_verify:
            if any(key not in load for key in ("id", "tgt", "fun")):
                return {}

        if isinstance(load["fun"], str):
            functions = list(set(load["fun"].split(",")))
            _ret_dict = len(functions) > 1
        elif isinstance(load["fun"], list):
            functions = load["fun"]
            _ret_dict = True
        else:
            return {}

        functions_allowed = []

        if "mine_get" in self.opts:
            # If master side acl defined.
            if not isinstance(self.opts["mine_get"], dict):
                return {}
            perms = set()
            for match in self.opts["mine_get"]:
                if re.match(match, load["id"]):
                    if isinstance(self.opts["mine_get"][match], list):
                        perms.update(self.opts["mine_get"][match])
            for fun in functions:
                if any(re.match(perm, fun) for perm in perms):
                    functions_allowed.append(fun)
            if not functions_allowed:
                return {}
        else:
            functions_allowed = functions

        ret = {}
        if not salt.utils.verify.valid_id(self.opts, load["id"]):
            return ret

        expr_form = load.get("expr_form")
        # keep both expr_form and tgt_type to ensure
        # comptability between old versions of salt
        if expr_form is not None and "tgt_type" not in load:
            match_type = expr_form
        else:
            match_type = load.get("tgt_type", "glob")
        if match_type.lower() == "pillar":
            match_type = "pillar_exact"
        if match_type.lower() == "compound":
            match_type = "compound_pillar_exact"
        checker = salt.utils.minions.CkMinions(self.opts)
        _res = checker.check_minions(load["tgt"], match_type, greedy=False)
        minions = _res["minions"]
        minion_side_acl = {}  # Cache minion-side ACL
        for minion in minions:
            mine_data = self.cache.fetch(f"minions/{minion}", "mine")
            if not isinstance(mine_data, dict):
                continue
            for function in functions_allowed:
                if function not in mine_data:
                    continue
                mine_entry = mine_data[function]
                mine_result = mine_data[function]
                if (
                    isinstance(mine_entry, dict)
                    and salt.utils.mine.MINE_ITEM_ACL_ID in mine_entry
                ):
                    mine_result = mine_entry[salt.utils.mine.MINE_ITEM_ACL_DATA]
                    # Check and fill minion-side ACL cache
                    if function not in minion_side_acl.get(minion, {}):
                        if "allow_tgt" in mine_entry:
                            # Only determine allowed targets if any have been specified.
                            # This prevents having to add a list of all minions as allowed targets.
                            get_minion = checker.check_minions(
                                mine_entry["allow_tgt"],
                                mine_entry.get("allow_tgt_type", "glob"),
                            )["minions"]
                            # the minion in allow_tgt does not exist
                            if not get_minion:
                                continue
                            salt.utils.dictupdate.set_dict_key_value(
                                minion_side_acl,
                                f"{minion}:{function}",
                                get_minion,
                            )
                if salt.utils.mine.minion_side_acl_denied(
                    minion_side_acl, minion, function, load["id"]
                ):
                    continue
                if _ret_dict:
                    ret.setdefault(function, {})[minion] = mine_result
                else:
                    # There is only one function in functions_allowed.
                    ret[minion] = mine_result
        return ret

    def _mine(self, load, skip_verify=False):
        """
        Store/update the mine data in cache.
        """
        if not skip_verify:
            if "id" not in load or "data" not in load:
                return False
        if self.opts.get("minion_data_cache", False) or self.opts.get(
            "enforce_mine_cache", False
        ):
            cbank = "minions/{}".format(load["id"])
            ckey = "mine"
            new_data = load["data"]
            if not load.get("clear", False):
                data = self.cache.fetch(cbank, ckey)
                if isinstance(data, dict):
                    data.update(new_data)
            self.cache.store(cbank, ckey, data)
        return True

    def _mine_delete(self, load):
        """
        Allow the minion to delete a specific function from its own mine
        """
        if "id" not in load or "fun" not in load:
            return False
        if self.opts.get("minion_data_cache", False) or self.opts.get(
            "enforce_mine_cache", False
        ):
            cbank = "minions/{}".format(load["id"])
            ckey = "mine"
            try:
                data = self.cache.fetch(cbank, ckey)
                if not isinstance(data, dict):
                    return False
                if load["fun"] in data:
                    del data[load["fun"]]
                    self.cache.store(cbank, ckey, data)
            except OSError:
                return False
        return True

    def _mine_flush(self, load, skip_verify=False):
        """
        Allow the minion to delete all of its own mine contents
        """
        if not skip_verify and "id" not in load:
            return False
        if self.opts.get("minion_data_cache", False) or self.opts.get(
            "enforce_mine_cache", False
        ):
            return self.cache.flush("minions/{}".format(load["id"]), "mine")
        return True

    def _file_recv(self, load):
        """
        Allows minions to send files to the master, files are sent to the
        master file cache
        """
        if any(key not in load for key in ("id", "path", "loc")):
            return False
        if not self.opts["file_recv"] or os.path.isabs(load["path"]):
            return False
        if os.path.isabs(load["path"]) or "../" in load["path"]:
            # Can overwrite master files!!
            return False
        if not salt.utils.verify.valid_id(self.opts, load["id"]):
            return False
        file_recv_max_size = 1024 * 1024 * self.opts["file_recv_max_size"]

        if "loc" in load and load["loc"] < 0:
            log.error("Invalid file pointer: load[loc] < 0")
            return False

        if load.get("size", 0) > file_recv_max_size:
            log.error("Exceeding file_recv_max_size limit: %s", file_recv_max_size)
            return False

        if len(load["data"]) + load.get("loc", 0) > file_recv_max_size:
            log.error("Exceeding file_recv_max_size limit: %s", file_recv_max_size)
            return False
        # Normalize Windows paths
        normpath = load["path"]
        if ":" in normpath:
            # make sure double backslashes are normalized
            normpath = normpath.replace("\\", "/")
            normpath = os.path.normpath(normpath)
        cpath = os.path.join(
            self.opts["cachedir"], "minions", load["id"], "files", normpath
        )
        cdir = os.path.dirname(cpath)
        if not os.path.isdir(cdir):
            try:
                os.makedirs(cdir)
            except OSError:
                pass
        if os.path.isfile(cpath) and load["loc"] != 0:
            mode = "ab"
        else:
            mode = "wb"
        with salt.utils.files.fopen(cpath, mode) as fp_:
            if load["loc"]:
                fp_.seek(load["loc"])
            fp_.write(salt.utils.stringutils.to_str(load["data"]))
        return True

    def _pillar(self, load):
        """
        Return the pillar data for the minion
        """
        if any(key not in load for key in ("id", "grains")):
            return False
        log.debug("Master _pillar using ext: %s", load.get("ext"))
        pillar = salt.pillar.get_pillar(
            self.opts,
            load["grains"],
            load["id"],
            load.get("saltenv", load.get("env")),
            load.get("ext"),
            self.mminion.functions,
            pillar_override=load.get("pillar_override", {}),
        )
        data = pillar.compile_pillar()
        if self.opts.get("minion_data_cache", False):
            self.cache.store(
                "minions/{}".format(load["id"]),
                "data",
                {"grains": load["grains"], "pillar": data},
            )
            if self.opts.get("minion_data_cache_events") is True:
                self.event.fire_event(
                    {"comment": "Minion data cache refresh"},
                    salt.utils.event.tagify(load["id"], "refresh", "minion"),
                )
        return data

    def _minion_event(self, load):
        """
        Receive an event from the minion and fire it on the master event
        interface
        """
        if "id" not in load:
            return False
        if "events" not in load and ("tag" not in load or "data" not in load):
            return False
        if "events" in load:
            for event in load["events"]:
                if "data" in event:
                    event_data = event["data"]
                else:
                    event_data = event
                self.event.fire_event(event_data, event["tag"])  # old dup event
                if load.get("pretag") is not None:
                    self.event.fire_event(
                        event_data,
                        salt.utils.event.tagify(event["tag"], base=load["pretag"]),
                    )
        else:
            tag = load["tag"]
            self.event.fire_event(load, tag)
        return True

    def _return(self, load):
        """
        Handle the return data sent from the minions
        """
        # Generate EndTime
        endtime = salt.utils.jid.jid_to_time(salt.utils.jid.gen_jid(self.opts))
        # If the return data is invalid, just ignore it
        if any(key not in load for key in ("return", "jid", "id")):
            return False

        if load["jid"] == "req":
            # The minion is returning a standalone job, request a jobid
            prep_fstr = "{}.prep_jid".format(self.opts["master_job_cache"])
            load["jid"] = self.mminion.returners[prep_fstr](
                nocache=load.get("nocache", False)
            )

            # save the load, since we don't have it
            saveload_fstr = "{}.save_load".format(self.opts["master_job_cache"])
            self.mminion.returners[saveload_fstr](load["jid"], load)
        log.info("Got return from %s for job %s", load["id"], load["jid"])
        self.event.fire_event(load, load["jid"])  # old dup event
        self.event.fire_event(
            load, salt.utils.event.tagify([load["jid"], "ret", load["id"]], "job")
        )
        self.event.fire_ret_load(load)
        if not self.opts["job_cache"] or self.opts.get("ext_job_cache"):
            return

        fstr = "{}.update_endtime".format(self.opts["master_job_cache"])
        if self.opts.get("job_cache_store_endtime") and fstr in self.mminion.returners:
            self.mminion.returners[fstr](load["jid"], endtime)

        fstr = "{}.returner".format(self.opts["master_job_cache"])
        self.mminion.returners[fstr](load)

    def _syndic_return(self, load):
        """
        Receive a syndic minion return and format it to look like returns from
        individual minions.
        """
        # Verify the load
        if any(key not in load for key in ("return", "jid", "id")):
            return None
        # if we have a load, save it
        if "load" in load:
            fstr = "{}.save_load".format(self.opts["master_job_cache"])
            self.mminion.returners[fstr](load["jid"], load["load"])

        # Format individual return loads
        for key, item in load["return"].items():
            ret = {"jid": load["jid"], "id": key, "return": item}
            if "out" in load:
                ret["out"] = load["out"]
            self._return(ret)

    def minion_runner(self, load):
        """
        Execute a runner from a minion, return the runner's function data
        """
        if "peer_run" not in self.opts:
            return {}
        if not isinstance(self.opts["peer_run"], dict):
            return {}
        if any(key not in load for key in ("fun", "arg", "id")):
            return {}
        perms = set()
        for match in self.opts["peer_run"]:
            if re.match(match, load["id"]):
                # This is the list of funcs/modules!
                if isinstance(self.opts["peer_run"][match], list):
                    perms.update(self.opts["peer_run"][match])
        good = False
        for perm in perms:
            if re.match(perm, load["fun"]):
                good = True
        if not good:
            # The minion is not who it says it is!
            # We don't want to listen to it!
            log.warning("Minion id %s is not who it says it is!", load["id"])
            return {}
        # Prepare the runner object
        opts = {}
        opts.update(self.opts)
        opts.update(
            {
                "fun": load["fun"],
                "arg": salt.utils.args.parse_input(
                    load["arg"], no_parse=load.get("no_parse", [])
                ),
                "id": load["id"],
                "doc": False,
                "conf_file": self.opts["conf_file"],
            }
        )
        runner = salt.runner.Runner(opts)
        return runner.run()

    def pub_ret(self, load, skip_verify=False):
        """
        Request the return data from a specific jid, only allowed
        if the requesting minion also initialted the execution.
        """
        if not skip_verify and any(key not in load for key in ("jid", "id")):
            return {}
        else:
            auth_cache = os.path.join(self.opts["cachedir"], "publish_auth")
            if not os.path.isdir(auth_cache):
                os.makedirs(auth_cache)
            jid_fn = os.path.join(auth_cache, load["jid"])
            with salt.utils.files.fopen(jid_fn, "r") as fp_:
                if not load["id"] == salt.utils.stringutils.to_unicode(fp_.read()):
                    return {}

            return self.local.get_cache_returns(load["jid"])

    def minion_pub(self, load):
        """
        Publish a command initiated from a minion, this method executes minion
        restrictions so that the minion publication will only work if it is
        enabled in the config.
        The configuration on the master allows minions to be matched to
        salt functions, so the minions can only publish allowed salt functions
        The config will look like this:
        peer:
            .*:
                - .*
        This configuration will enable all minions to execute all commands.
        peer:
            foo.example.com:
                - test.*
        This configuration will only allow the minion foo.example.com to
        execute commands from the test module
        """
        if not self.__verify_minion_publish(load):
            return {}
        # Set up the publication payload
        pub_load = {
            "fun": load["fun"],
            "arg": salt.utils.args.parse_input(
                load["arg"], no_parse=load.get("no_parse", [])
            ),
            "tgt_type": load.get("tgt_type", "glob"),
            "tgt": load["tgt"],
            "ret": load["ret"],
            "id": load["id"],
        }
        if "tgt_type" in load:
            if load["tgt_type"].startswith("node"):
                if load["tgt"] in self.opts["nodegroups"]:
                    pub_load["tgt"] = self.opts["nodegroups"][load["tgt"]]
                    pub_load["tgt_type"] = "compound"
                else:
                    return {}
            else:
                pub_load["tgt_type"] = load["tgt_type"]
        ret = {}
        ret["jid"] = self.local.cmd_async(**pub_load)
        _res = self.ckminions.check_minions(load["tgt"], pub_load["tgt_type"])
        ret["minions"] = _res["minions"]
        auth_cache = os.path.join(self.opts["cachedir"], "publish_auth")
        if not os.path.isdir(auth_cache):
            os.makedirs(auth_cache)
        jid_fn = os.path.join(auth_cache, str(ret["jid"]))
        with salt.utils.files.fopen(jid_fn, "w+") as fp_:
            fp_.write(salt.utils.stringutils.to_str(load["id"]))
        return ret

    def minion_publish(self, load):
        """
        Publish a command initiated from a minion, this method executes minion
        restrictions so that the minion publication will only work if it is
        enabled in the config.
        The configuration on the master allows minions to be matched to
        salt functions, so the minions can only publish allowed salt functions
        The config will look like this:
        peer:
            .*:
                - .*
        This configuration will enable all minions to execute all commands.
        peer:
            foo.example.com:
                - test.*
        This configuration will only allow the minion foo.example.com to
        execute commands from the test module
        """
        if not self.__verify_minion_publish(load):
            return {}
        # Set up the publication payload
        pub_load = {
            "fun": load["fun"],
            "arg": salt.utils.args.parse_input(
                load["arg"], no_parse=load.get("no_parse", [])
            ),
            "tgt_type": load.get("tgt_type", "glob"),
            "tgt": load["tgt"],
            "ret": load["ret"],
            "id": load["id"],
        }
        if "tmo" in load:
            try:
                pub_load["timeout"] = int(load["tmo"])
            except ValueError:
                msg = "Failed to parse timeout value: {}".format(load["tmo"])
                log.warning(msg)
                return {}
        if "timeout" in load:
            try:
                pub_load["timeout"] = int(load["timeout"])
            except ValueError:
                msg = "Failed to parse timeout value: {}".format(load["timeout"])
                log.warning(msg)
                return {}
        if "tgt_type" in load:
            if load["tgt_type"].startswith("node"):
                if load["tgt"] in self.opts["nodegroups"]:
                    pub_load["tgt"] = self.opts["nodegroups"][load["tgt"]]
                    pub_load["tgt_type"] = "compound"
                else:
                    return {}
            else:
                pub_load["tgt_type"] = load["tgt_type"]
        pub_load["raw"] = True
        ret = {}
        for minion in self.local.cmd_iter(**pub_load):
            if load.get("form", "") == "full":
                data = minion
                if "jid" in minion:
                    ret["__jid__"] = minion["jid"]
                data["ret"] = data.pop("return")
                ret[minion["id"]] = data
            else:
                ret[minion["id"]] = minion["return"]
                if "jid" in minion:
                    ret["__jid__"] = minion["jid"]
        for key, val in self.local.get_cache_returns(ret["__jid__"]).items():
            if key not in ret:
                ret[key] = val
        if load.get("form", "") != "full":
            ret.pop("__jid__")
        return ret

    def revoke_auth(self, load):
        """
        Allow a minion to request revocation of its own key
        """
        if "id" not in load:
            return False
        with salt.key.Key(self.opts) as keyapi:
            keyapi.delete_key(
                load["id"], preserve_minions=load.get("preserve_minion_cache", False)
            )
        return True

    def destroy(self):
        if self.event is not None:
            self.event.destroy()
            self.event = None
        if self.local is not None:
            self.local.destroy()
            self.local = None


class LocalFuncs:
    """
    Set up methods for use only from the local system
    """

    # The ClearFuncs object encapsulates the functions that can be executed in
    # the clear:
    # publish (The publish from the LocalClient)
    # _auth
    def __init__(self, opts, key):
        self.opts = opts
        self.key = key
        # Create the event manager
        self.event = salt.utils.event.get_event(
            "master",
            self.opts["sock_dir"],
            opts=self.opts,
            listen=False,
        )
        # Make a client
        self.local = salt.client.get_local_client(mopts=self.opts)
        # Make an minion checker object
        self.ckminions = salt.utils.minions.CkMinions(opts)
        # Make an Auth object
        self.loadauth = salt.auth.LoadAuth(opts)
        # Stand up the master Minion to access returner data
        self.mminion = salt.minion.MasterMinion(self.opts, states=False, rend=False)
        # Make a wheel object
        self.wheel_ = salt.wheel.Wheel(opts)

    def runner(self, load):
        """
        Send a master control function back to the runner system
        """
        # All runner opts pass through eauth
        auth_type, err_name, key = self._prep_auth_info(load)

        # Authenticate
        auth_check = self.loadauth.check_authentication(load, auth_type)
        error = auth_check.get("error")

        if error:
            # Authentication error occurred: do not continue.
            return {"error": error}

        # Authorize
        runner_check = self.ckminions.runner_check(
            auth_check.get("auth_list", []), load["fun"], load["kwarg"]
        )
        username = auth_check.get("username")
        if not runner_check:
            return {
                "error": {
                    "name": err_name,
                    "message": (
                        'Authentication failure of type "{}" occurred '
                        "for user {}.".format(auth_type, username)
                    ),
                }
            }
        elif isinstance(runner_check, dict) and "error" in runner_check:
            # A dictionary with an error name/message was handled by ckminions.runner_check
            return runner_check

        # Authorized. Do the job!
        try:
            fun = load.pop("fun")
            runner_client = salt.runner.RunnerClient(self.opts)
            return runner_client.asynchronous(fun, load.get("kwarg", {}), username)
        except Exception as exc:  # pylint: disable=broad-except
            log.exception("Exception occurred while introspecting %s")
            return {
                "error": {
                    "name": exc.__class__.__name__,
                    "args": exc.args,
                    "message": str(exc),
                }
            }

    def wheel(self, load):
        """
        Send a master control function back to the wheel system
        """
        # All wheel ops pass through eauth
        auth_type, err_name, key = self._prep_auth_info(load)

        # Authenticate
        auth_check = self.loadauth.check_authentication(
            load, auth_type, key=key, show_username=True
        )
        error = auth_check.get("error")

        if error:
            # Authentication error occurred: do not continue.
            return {"error": error}

        # Authorize
        username = auth_check.get("username")
        if auth_type != "user":
            wheel_check = self.ckminions.wheel_check(
                auth_check.get("auth_list", []), load["fun"], load["kwarg"]
            )
            if not wheel_check:
                return {
                    "error": {
                        "name": err_name,
                        "message": (
                            'Authentication failure of type "{}" occurred for '
                            "user {}.".format(auth_type, username)
                        ),
                    }
                }
            elif isinstance(wheel_check, dict) and "error" in wheel_check:
                # A dictionary with an error name/message was handled by ckminions.wheel_check
                return wheel_check

        # Authenticated. Do the job.
        jid = salt.utils.jid.gen_jid(self.opts)
        fun = load.pop("fun")
        tag = salt.utils.event.tagify(jid, prefix="wheel")
        data = {
            "fun": f"wheel.{fun}",
            "jid": jid,
            "tag": tag,
            "user": username,
        }
        try:
            self.event.fire_event(data, salt.utils.event.tagify([jid, "new"], "wheel"))
            ret = self.wheel_.call_func(fun, **load)
            data["return"] = ret
            data["success"] = True
            self.event.fire_event(data, salt.utils.event.tagify([jid, "ret"], "wheel"))
            return {"tag": tag, "data": data}
        except Exception as exc:  # pylint: disable=broad-except
            log.exception("Exception occurred while introspecting %s", fun)
            data["return"] = "Exception occurred in wheel {}: {}: {}".format(
                fun,
                exc.__class__.__name__,
                exc,
            )
            data["success"] = False
            self.event.fire_event(data, salt.utils.event.tagify([jid, "ret"], "wheel"))
            return {"tag": tag, "data": data}

    def mk_token(self, load):
        """
        Create and return an authentication token, the clear load needs to
        contain the eauth key and the needed authentication creds.
        """
        token = self.loadauth.mk_token(load)
        if not token:
            log.warning('Authentication failure of type "eauth" occurred.')
            return ""
        return token

    def get_token(self, load):
        """
        Return the name associated with a token or False if the token is invalid
        """
        if "token" not in load:
            return False
        return self.loadauth.get_tok(load["token"])

    def publish(self, load):
        """
        This method sends out publications to the minions, it can only be used
        by the LocalClient.
        """
        extra = load.get("kwargs", {})

        publisher_acl = salt.acl.PublisherACL(self.opts["publisher_acl_blacklist"])

        if publisher_acl.user_is_blacklisted(
            load["user"]
        ) or publisher_acl.cmd_is_blacklisted(load["fun"]):
            log.error(
                "%s does not have permissions to run %s. Please contact "
                "your local administrator if you believe this is in error.",
                load["user"],
                load["fun"],
            )
            return {
                "error": {
                    "name": "AuthorizationError",
                    "message": "Authorization error occurred.",
                }
            }

        # Retrieve the minions list
        delimiter = load.get("kwargs", {}).get("delimiter", DEFAULT_TARGET_DELIM)
        _res = self.ckminions.check_minions(
            load["tgt"], load.get("tgt_type", "glob"), delimiter
        )
        minions = _res["minions"]

        # Check for external auth calls and authenticate
        auth_type, err_name, key = self._prep_auth_info(extra)
        if auth_type == "user":
            auth_check = self.loadauth.check_authentication(load, auth_type, key=key)
        else:
            auth_check = self.loadauth.check_authentication(extra, auth_type)

        # Setup authorization list variable and error information
        auth_list = auth_check.get("auth_list", [])
        error = auth_check.get("error")
        err_msg = f'Authentication failure of type "{auth_type}" occurred.'

        if error:
            # Authentication error occurred: do not continue.
            log.warning(err_msg)
            return {
                "error": {
                    "name": "AuthenticationError",
                    "message": "Authentication error occurred.",
                }
            }

        # All Token, Eauth, and non-root users must pass the authorization check
        if auth_type != "user" or (auth_type == "user" and auth_list):
            # Authorize the request
            authorized = self.ckminions.auth_check(
                auth_list,
                load["fun"],
                load["arg"],
                load["tgt"],
                load.get("tgt_type", "glob"),
                minions=minions,
                # always accept find_job
                whitelist=["saltutil.find_job"],
            )

            if not authorized:
                # Authorization error occurred. Log warning and do not continue.
                log.warning(err_msg)
                return {
                    "error": {
                        "name": "AuthorizationError",
                        "message": "Authorization error occurred.",
                    }
                }

            # Perform some specific auth_type tasks after the authorization check
            if auth_type == "token":
                username = auth_check.get("username")
                load["user"] = username
                log.debug('Minion tokenized user = "%s"', username)
            elif auth_type == "eauth":
                # The username we are attempting to auth with
                load["user"] = self.loadauth.load_name(extra)

        # If we order masters (via a syndic), don't short circuit if no minions
        # are found
        if not self.opts.get("order_masters"):
            # Check for no minions
            if not minions:
                return {"enc": "clear", "load": {"jid": None, "minions": minions}}
        # Retrieve the jid
        if not load["jid"]:
            fstr = "{}.prep_jid".format(self.opts["master_job_cache"])
            load["jid"] = self.mminion.returners[fstr](
                nocache=extra.get("nocache", False)
            )
        self.event.fire_event({"minions": minions}, load["jid"])

        new_job_load = {
            "jid": load["jid"],
            "tgt_type": load["tgt_type"],
            "tgt": load["tgt"],
            "user": load["user"],
            "fun": load["fun"],
            "arg": salt.utils.args.parse_input(
                load["arg"], no_parse=load.get("no_parse", [])
            ),
            "minions": minions,
        }

        # Announce the job on the event bus
        self.event.fire_event(new_job_load, "new_job")  # old dup event
        self.event.fire_event(
            new_job_load, salt.utils.event.tagify([load["jid"], "new"], "job")
        )

        # Save the invocation information
        if self.opts["ext_job_cache"]:
            try:
                fstr = "{}.save_load".format(self.opts["ext_job_cache"])
                self.mminion.returners[fstr](load["jid"], load)
            except KeyError:
                log.critical(
                    "The specified returner used for the external job cache "
                    '"%s" does not have a save_load function!',
                    self.opts["ext_job_cache"],
                )
            except Exception:  # pylint: disable=broad-except
                log.critical(
                    "The specified returner threw a stack trace:", exc_info=True
                )

        # always write out to the master job cache
        try:
            fstr = "{}.save_load".format(self.opts["master_job_cache"])
            self.mminion.returners[fstr](load["jid"], load)
        except KeyError:
            log.critical(
                "The specified returner used for the master job cache "
                '"%s" does not have a save_load function!',
                self.opts["master_job_cache"],
            )
        except Exception:  # pylint: disable=broad-except
            log.critical("The specified returner threw a stack trace:", exc_info=True)
        # Altering the contents of the publish load is serious!! Changes here
        # break compatibility with minion/master versions and even tiny
        # additions can have serious implications on the performance of the
        # publish commands.
        #
        # In short, check with Thomas Hatch before you even think about
        # touching this stuff, we can probably do what you want to do another
        # way that won't have a negative impact.
        pub_load = {
            "fun": load["fun"],
            "arg": salt.utils.args.parse_input(
                load["arg"], no_parse=load.get("no_parse", [])
            ),
            "tgt": load["tgt"],
            "jid": load["jid"],
            "ret": load["ret"],
        }

        if "id" in extra:
            pub_load["id"] = extra["id"]
        if "tgt_type" in load:
            pub_load["tgt_type"] = load["tgt_type"]
        if "to" in load:
            pub_load["to"] = load["to"]

        if "kwargs" in load:
            if "ret_config" in load["kwargs"]:
                pub_load["ret_config"] = load["kwargs"].get("ret_config")

            if "metadata" in load["kwargs"]:
                pub_load["metadata"] = load["kwargs"].get("metadata")

            if "ret_kwargs" in load["kwargs"]:
                pub_load["ret_kwargs"] = load["kwargs"].get("ret_kwargs")

        if "user" in load:
            log.info(
                "User %s Published command %s with jid %s",
                load["user"],
                load["fun"],
                load["jid"],
            )
            pub_load["user"] = load["user"]
        else:
            log.info("Published command %s with jid %s", load["fun"], load["jid"])
        log.debug("Published command details %s", pub_load)

        return {"ret": {"jid": load["jid"], "minions": minions}, "pub": pub_load}

    def _prep_auth_info(self, load):
        key = None
        if "token" in load:
            auth_type = "token"
            err_name = "TokenAuthenticationError"
        elif "eauth" in load:
            auth_type = "eauth"
            err_name = "EauthAuthenticationError"
        else:
            auth_type = "user"
            err_name = "UserAuthenticationError"
            key = self.key

        return auth_type, err_name, key

    def destroy(self):
        if self.event is not None:
            self.event.destroy()
            self.event = None
        if self.local is not None:
            self.local.destroy()
            self.local = None



import pytest

import salt.pillar.saltclass as saltclass


@pytest.fixture
def configure_loader_modules():
    return {saltclass: {}}


@pytest.fixture
def minion_id():
    return "fake_id"


@pytest.fixture
def temp_saltclass_tree(tmp_path, minion_id):
    dirname = tmp_path / "saltclass" / "examples"
    dirname.mkdir(parents=True, exist_ok=True)
    classes_dir = dirname / "classes"
    classes_dir.mkdir(parents=True, exist_ok=True)
    nodes_dir = dirname / "nodes"
    nodes_dir.mkdir(parents=True, exist_ok=True)
    default_dir = classes_dir / "default"
    default_dir.mkdir(parents=True, exist_ok=True)
    users_dir = default_dir / "users"
    users_dir.mkdir(parents=True, exist_ok=True)
    roles_dir = classes_dir / "roles"
    roles_dir.mkdir(parents=True, exist_ok=True)
    nginx_subdir = roles_dir / "nginx"
    nginx_subdir.mkdir(parents=True, exist_ok=True)

    default_init = default_dir / "init.yml"
    test_list = """
    classes:
      - default.users
      - default.motd
      - default.empty

    states:
      - default

    pillars:
      default:
        network:
          dns:
    {% if __grains__['os'] == 'should_never_match' %}
            srv1: 192.168.0.1
            srv2: 192.168.0.2
            domain: example.com
    {% endif %}
          ntp:
            srv1: 192.168.10.10
            srv2: 192.168.10.20
      test_list:
        - a: ${default:network:ntp:srv1}
        - ${default:network:ntp:srv2}

      global_scalar: from default
      test_dict:
        a_scalar: from default
        a_list:
          - element1
          - element2
    """
    default_init.write_text(test_list)

    minion_node_file = nodes_dir / f"{minion_id}.yml"
    nodes_text = """
    environment: base

    states:
      - minion_node

    classes:
    {% for class in ['default', 'roles.*', 'empty.*'] %}
      - {{ class }}
    {% endfor %}

    pillars:
      global_scalar: from minion_node
      test_dict:
        a_scalar: from minion_node
    """
    minion_node_file.write_text(nodes_text)

    (users_dir / "init.yml").write_text(
        """
        classes:
          - default.users.foo

        states:
          - users

        pillars:
          default:
            ntp:
              srv1: 192.168.20.10

          global_scalar: from users
          test_dict:
            a_scalar: from users
        """
    )
    (users_dir / "foo.yml").write_text(
        """
        states:
          - users.foo

        global_scalar: from users.foo
        users_foo_scalar: from users.foo
        test_dict:
          a_scalar: from users.foo
        """
    )
    (default_dir / "empty.yml").write_text("test: this is a test")
    (default_dir / "motd.yml").write_text(
        """
        states:
          - motd

        pillars:
          global_scalar: from motd
        """
    )
    (roles_dir / "app.yml").write_text(
        """
        states:
          - app

        pillars:
          global_scalar: from app
        """
    )
    (nginx_subdir / "init.yml").write_text(
        """
        states:
          - nginx

        pillars:
          global_scalar: from nginx
          nginx_scalar: from nginx
        """
    )

    return dirname


def test_classes_order(temp_saltclass_tree):
    """
    Classes must be correctly ordered.

    See: https://github.com/saltstack/salt/issues/58969
    """
    expected_ret = [
        "default.users.foo",
        "default.users",
        "default.motd",
        "default.empty",
        "default",
        "roles.app",
        "roles.nginx",
    ]
    fake_args = {"path": str(temp_saltclass_tree)}
    fake_pillar = {}
    fake_minion_id = "fake_id"
    try:
        full_ret = saltclass.ext_pillar(fake_minion_id, fake_pillar, fake_args)
        parsed_ret = full_ret["__saltclass__"]["classes"]
    # Fail the test if we hit our NoneType error
    except TypeError as err:
        pytest.fail(err)
    # Else give the parsed content result
    assert expected_ret == parsed_ret


def test_list_expansion_succeeds(temp_saltclass_tree):
    expected_ret = [{"a": "192.168.10.10"}, "192.168.10.20"]
    full_ret = {}
    parsed_ret = []
    fake_args = {"path": str(temp_saltclass_tree)}
    fake_pillar = {}
    fake_minion_id = "fake_id"
    try:
        full_ret = saltclass.ext_pillar(fake_minion_id, fake_pillar, fake_args)
        parsed_ret = full_ret["test_list"]
    # Fail the test if we hit our NoneType error
    except TypeError as err:
        pytest.fail(err)
    # Else give the parsed content result
    assert expected_ret == parsed_ret


def test_node_override_classes_scalars(temp_saltclass_tree):
    """
    Scalars pillars defined in a node definition must override the
    definition from classes.
    """
    expected_ret = "from minion_node"
    fake_args = {"path": str(temp_saltclass_tree)}
    fake_pillar = {}
    fake_minion_id = "fake_id"
    try:
        full_ret = saltclass.ext_pillar(fake_minion_id, fake_pillar, fake_args)
        parsed_ret = full_ret["global_scalar"]
    # Fail the test if we hit our NoneType error
    except TypeError as err:
        pytest.fail(err)
    # Else give the parsed content result
    assert expected_ret == parsed_ret


def test_node_override_classes_scalar_in_dict(temp_saltclass_tree):
    """
    Scalars defined in `dict` pillars defined in a node definition must override the
    same dict definition from classes.

    See: https://github.com/saltstack/salt/issues/63933
    """
    expected_ret = "from minion_node"
    fake_args = {"path": str(temp_saltclass_tree)}
    fake_pillar = {}
    fake_minion_id = "fake_id"
    try:
        full_ret = saltclass.ext_pillar(fake_minion_id, fake_pillar, fake_args)
        parsed_ret = full_ret["test_dict"]["a_scalar"]
    # Fail the test if we hit our NoneType error
    except TypeError as err:
        pytest.fail(err)
    # Else give the parsed content result
    assert expected_ret == parsed_ret


def test_node_override_classes_list_in_dict(temp_saltclass_tree):
    """
    `list` under a `dict` defined in a node definition must override the
    same definition from classes.

    See: https://github.com/saltstack/salt/issues/63933
    """
    expected_ret = {"srv1": "192.168.10.10", "srv2": "192.168.10.20"}
    fake_args = {"path": str(temp_saltclass_tree)}
    fake_pillar = {}
    fake_minion_id = "fake_id"
    try:
        full_ret = saltclass.ext_pillar(fake_minion_id, fake_pillar, fake_args)
        parsed_ret = full_ret["default"]["network"]["ntp"]
    # Fail the test if we hit our NoneType error
    except TypeError as err:
        pytest.fail(err)
    # Else give the parsed content result
    assert expected_ret == parsed_ret


def test_list_in_dict_no_duplication(temp_saltclass_tree):
    """
    `list` under a `dict` in pillar must not be duplicated.

    See:
    """
    expected_ret = ["element1", "element2"]
    fake_args = {"path": str(temp_saltclass_tree)}
    fake_pillar = {}
    fake_minion_id = "fake_id"
    try:
        full_ret = saltclass.ext_pillar(fake_minion_id, fake_pillar, fake_args)
        parsed_ret = full_ret["test_dict"]["a_list"]
    # Fail the test if we hit our NoneType error
    except TypeError as err:
        pytest.fail(err)
    # Else give the parsed content result
    assert expected_ret == parsed_ret


def test_nested_classes_has_pillars(temp_saltclass_tree):
    """
    pillars defined in nested classes are present
    """
    expected_ret = "from nginx"
    fake_args = {"path": str(temp_saltclass_tree)}
    fake_pillar = {}
    fake_minion_id = "fake_id"
    try:
        full_ret = saltclass.ext_pillar(fake_minion_id, fake_pillar, fake_args)
        parsed_ret = full_ret["nginx_scalar"]
    # Fail the test if we hit our NoneType error
    except TypeError as err:
        pytest.fail(err)
    # Else give the parsed content result
    assert expected_ret == parsed_ret



"""
For running command line executables with a timeout
"""

import shlex
import subprocess
import threading

import salt.exceptions
import salt.utils.data
import salt.utils.platform
import salt.utils.stringutils


class TimedProc:
    """
    Create a TimedProc object, calls subprocess.Popen with passed args and **kwargs
    """

    def __init__(self, args, **kwargs):

        self.wait = not kwargs.pop("bg", False)
        self.stdin = kwargs.pop("stdin", None)
        self.with_communicate = kwargs.pop("with_communicate", self.wait)
        self.timeout = kwargs.pop("timeout", None)
        self.stdin_raw_newlines = kwargs.pop("stdin_raw_newlines", False)

        # If you're not willing to wait for the process
        # you can't define any stdin, stdout or stderr
        if not self.wait:
            self.stdin = kwargs["stdin"] = None
            self.with_communicate = False
        elif self.stdin is not None:
            if not self.stdin_raw_newlines:
                # Translate a newline submitted as '\n' on the CLI to an actual
                # newline character.
                self.stdin = self.stdin.replace("\\n", "\n")
            self.stdin = salt.utils.stringutils.to_bytes(self.stdin)
            kwargs["stdin"] = subprocess.PIPE

        if not self.with_communicate:
            self.stdout = kwargs["stdout"] = None
            self.stderr = kwargs["stderr"] = None

        if self.timeout and not isinstance(self.timeout, (int, float)):
            raise salt.exceptions.TimedProcTimeoutError(
                f"Error: timeout {self.timeout} must be a number"
            )
        if kwargs.get("shell", False):
            args = salt.utils.data.decode(args, to_str=True)

        try:
            self.process = subprocess.Popen(args, **kwargs)
        except (AttributeError, TypeError):
            if not kwargs.get("shell", False):
                use_posix = not salt.utils.platform.is_windows()
                if not isinstance(args, (list, tuple)):
                    try:
                        args = shlex.split(args, posix=use_posix)
                    except AttributeError:
                        args = shlex.split(str(args), posix=use_posix)
                str_args = []
                for arg in args:
                    if not isinstance(arg, str):
                        str_args.append(str(arg))
                    else:
                        str_args.append(arg)
                args = str_args
            else:
                if not isinstance(args, (list, tuple, str)):
                    # Handle corner case where someone does a 'cmd.run 3'
                    args = str(args)
            # Ensure that environment variables are strings
            for key, val in kwargs.get("env", {}).items():
                if not isinstance(val, str):
                    kwargs["env"][key] = str(val)
                if not isinstance(key, str):
                    kwargs["env"][str(key)] = kwargs["env"].pop(key)
            args = salt.utils.data.decode(args)
            self.process = subprocess.Popen(args, **kwargs)
        self.command = args

    def run(self):
        """
        wait for subprocess to terminate and return subprocess' return code.
        If timeout is reached, throw TimedProcTimeoutError
        """

        def receive():
            if self.with_communicate:
                self.stdout, self.stderr = self.process.communicate(input=self.stdin)
            elif self.wait:
                self.process.wait()

        if not self.timeout:
            receive()
        else:
            rt = threading.Thread(target=receive)
            rt.start()
            rt.join(self.timeout)
            if rt.is_alive():
                # Subprocess cleanup (best effort)
                self.process.kill()

                def terminate():
                    if rt.is_alive():
                        self.process.terminate()

                threading.Timer(10, terminate).start()
                raise salt.exceptions.TimedProcTimeoutError(
                    "{} : Timed out after {} seconds".format(
                        self.command,
                        str(self.timeout),
                    )
                )
        return self.process.returncode



import pytest
from tornado.httpclient import HTTPError

import salt.utils.json
from salt.netapi.rest_tornado import saltnado


@pytest.fixture
def app_urls(salt_sub_minion):
    return [
        ("/", saltnado.SaltAPIHandler),
    ]


async def test_root(http_client):
    """
    Test the root path which returns the list of clients we support
    """
    response = await http_client.fetch(
        "/", connect_timeout=30, request_timeout=30, headers=None
    )
    assert response.code == 200
    response_obj = salt.utils.json.loads(response.body)
    assert sorted(response_obj["clients"]) == [
        "local",
        "local_async",
        "runner",
        "runner_async",
    ]
    assert response_obj["return"] == "Welcome"


@pytest.mark.slow_test
async def test_post_no_auth(http_client, content_type_map):
    """
    Test post with no auth token, should 401
    """
    # get a token for this test
    low = [{"client": "local", "tgt": "*", "fun": "test.ping"}]
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/",
            method="POST",
            body=salt.utils.json.dumps(low),
            headers={"Content-Type": content_type_map["json"]},
            follow_redirects=False,
            connect_timeout=30,
            request_timeout=30,
        )
    assert exc.value.code == 302
    assert exc.value.response.headers["Location"] == "/login"


# Local client tests


async def test_simple_local_post(http_client, salt_minion, salt_sub_minion):
    """
    Test a basic API of /
    """
    low = [{"client": "local", "tgt": "*", "fun": "test.ping"}]
    response = await http_client.fetch(
        "/",
        method="POST",
        body=salt.utils.json.dumps(low),
        connect_timeout=30,
        request_timeout=30,
    )
    response_obj = salt.utils.json.loads(response.body)
    assert len(response_obj["return"]) == 1
    assert response_obj["return"][0] == {salt_minion.id: True, salt_sub_minion.id: True}


async def test_simple_local_post_no_tgt(http_client):
    """
    POST job with invalid tgt
    """
    low = [{"client": "local", "tgt": "minion_we_dont_have", "fun": "test.ping"}]
    response = await http_client.fetch(
        "/",
        method="POST",
        body=salt.utils.json.dumps(low),
        connect_timeout=30,
        request_timeout=30,
    )
    response_obj = salt.utils.json.loads(response.body)
    assert response_obj["return"] == [
        "No minions matched the target. No command was sent, no jid was assigned."
    ]


# local client request body test


async def test_simple_local_post_only_dictionary_request(
    http_client, salt_minion, salt_sub_minion
):
    """
    Test a basic API of /
    """
    low = {
        "client": "local",
        "tgt": "*",
        "fun": "test.ping",
    }
    response = await http_client.fetch(
        "/",
        method="POST",
        body=salt.utils.json.dumps(low),
        connect_timeout=30,
        request_timeout=30,
    )
    response_obj = salt.utils.json.loads(response.body)
    assert len(response_obj["return"]) == 1
    assert response_obj["return"][0] == {salt_minion.id: True, salt_sub_minion.id: True}


async def test_simple_local_post_invalid_request(http_client):
    """
    Test a basic API of /
    """
    low = ["invalid request"]
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/",
            method="POST",
            body=salt.utils.json.dumps(low),
            connect_timeout=30,
            request_timeout=30,
        )
    assert exc.value.code == 400


# local_async tests
async def test_simple_local_async_post(http_client, salt_minion, salt_sub_minion):
    low = [{"client": "local_async", "tgt": "*", "fun": "test.ping"}]
    response = await http_client.fetch(
        "/",
        method="POST",
        body=salt.utils.json.dumps(low),
    )

    response_obj = salt.utils.json.loads(response.body)
    ret = response_obj["return"]
    ret[0]["minions"] = sorted(ret[0]["minions"])

    # TODO: verify pub function? Maybe look at how we test the publisher
    assert len(ret) == 1
    assert "jid" in ret[0]
    assert ret[0]["minions"] == sorted([salt_minion.id, salt_sub_minion.id])


async def test_multi_local_async_post(http_client, salt_minion, salt_sub_minion):
    low = [
        {"client": "local_async", "tgt": "*", "fun": "test.ping"},
        {"client": "local_async", "tgt": "*", "fun": "test.ping"},
    ]
    response = await http_client.fetch(
        "/",
        method="POST",
        body=salt.utils.json.dumps(low),
    )

    response_obj = salt.utils.json.loads(response.body)
    ret = response_obj["return"]
    ret[0]["minions"] = sorted(ret[0]["minions"])
    ret[1]["minions"] = sorted(ret[1]["minions"])

    assert len(ret) == 2
    assert "jid" in ret[0]
    assert "jid" in ret[1]
    assert ret[0]["minions"] == sorted([salt_minion.id, salt_sub_minion.id])
    assert ret[1]["minions"] == sorted([salt_minion.id, salt_sub_minion.id])


@pytest.mark.slow_test
async def test_multi_local_async_post_multitoken(
    http_client, auth_token, salt_minion, salt_sub_minion
):
    low = [
        {"client": "local_async", "tgt": "*", "fun": "test.ping"},
        {
            "client": "local_async",
            "tgt": "*",
            "fun": "test.ping",
            # send a different (but still valid token)
            "token": auth_token["token"],
        },
        {
            "client": "local_async",
            "tgt": "*",
            "fun": "test.ping",
            "token": "BAD_TOKEN",  # send a bad token
        },
    ]
    response = await http_client.fetch(
        "/",
        method="POST",
        body=salt.utils.json.dumps(low),
    )

    response_obj = salt.utils.json.loads(response.body)
    ret = response_obj["return"]
    ret[0]["minions"] = sorted(ret[0]["minions"])
    ret[1]["minions"] = sorted(ret[1]["minions"])

    assert len(ret) == 3  # make sure we got 3 responses
    assert "jid" in ret[0]  # the first 2 are regular returns
    assert "jid" in ret[1]
    assert "Failed to authenticate" in ret[2]  # bad auth
    assert ret[0]["minions"] == sorted([salt_minion.id, salt_sub_minion.id])
    assert ret[1]["minions"] == sorted([salt_minion.id, salt_sub_minion.id])


@pytest.mark.slow_test
async def test_simple_local_async_post_no_tgt(http_client):
    low = [{"client": "local_async", "tgt": "minion_we_dont_have", "fun": "test.ping"}]
    response = await http_client.fetch(
        "/",
        method="POST",
        body=salt.utils.json.dumps(low),
    )
    response_obj = salt.utils.json.loads(response.body)
    assert response_obj["return"] == [{}]


# runner tests
@pytest.mark.slow_test
async def test_simple_local_runner_post(http_client, salt_minion, salt_sub_minion):
    low = [{"client": "runner", "fun": "manage.up"}]
    response = await http_client.fetch(
        "/",
        method="POST",
        body=salt.utils.json.dumps(low),
        connect_timeout=30,
        request_timeout=300,
    )
    response_obj = salt.utils.json.loads(response.body)
    assert len(response_obj["return"]) == 1
    assert sorted(response_obj["return"][0]) == sorted(
        [salt_minion.id, salt_sub_minion.id]
    )


# runner_async tests
async def test_simple_local_runner_async_post(http_client):
    low = [{"client": "runner_async", "fun": "manage.up"}]
    response = await http_client.fetch(
        "/",
        method="POST",
        body=salt.utils.json.dumps(low),
        connect_timeout=10,
        request_timeout=10,
    )
    response_obj = salt.utils.json.loads(response.body)
    assert "return" in response_obj
    assert 1 == len(response_obj["return"])
    assert "jid" in response_obj["return"][0]
    assert "tag" in response_obj["return"][0]



"""
Use the :ref:`Salt Event System <events>` to fire events from the
master to the minion and vice-versa.
"""

import logging
import os
import sys
import traceback
from collections.abc import Mapping

import salt.channel.client
import salt.crypt
import salt.payload
import salt.utils.event
import salt.utils.network

__proxyenabled__ = ["*"]
log = logging.getLogger(__name__)


def _dict_subset(keys, master_dict):
    """
    Return a dictionary of only the subset of keys/values specified in keys
    """
    return {k: v for k, v in master_dict.items() if k in keys}


def fire_master(data, tag, preload=None):
    """
    Fire an event off up to the master server

    CLI Example:

    .. code-block:: bash

        salt '*' event.fire_master '{"data":"my event data"}' 'tag'
    """
    if (
        __opts__.get("local", None) or __opts__.get("file_client", None) == "local"
    ) and not __opts__.get("use_master_when_local", False):
        #  We can't send an event if we're in masterless mode
        log.warning("Local mode detected. Event with tag %s will NOT be sent.", tag)
        return False

    if preload or __opts__.get("__cli") == "salt-call":
        # If preload is specified, we must send a raw event (this is
        # slower because it has to independently authenticate)
        if "master_uri" not in __opts__:
            __opts__["master_uri"] = "tcp://{ip}:{port}".format(
                ip=salt.utils.network.ip_bracket(__opts__["interface"]),
                port=__opts__.get("ret_port", "4506"),  # TODO, no fallback
            )
        masters = list()
        ret = True
        if "master_uri_list" in __opts__:
            for master_uri in __opts__["master_uri_list"]:
                masters.append(master_uri)
        else:
            masters.append(__opts__["master_uri"])
        auth = salt.crypt.SAuth(__opts__)
        load = {
            "id": __opts__["id"],
            "tag": tag,
            "data": data,
            "tok": auth.gen_token(b"salt"),
            "cmd": "_minion_event",
        }

        if isinstance(preload, dict):
            load.update(preload)

        for master in masters:
            with salt.channel.client.ReqChannel.factory(
                __opts__, master_uri=master
            ) as channel:
                try:
                    channel.send(load)
                    # channel.send was successful.
                    # Ensure ret is True.
                    ret = True
                except Exception:  # pylint: disable=broad-except
                    ret = False
        return ret
    else:
        # Usually, we can send the event via the minion, which is faster
        # because it is already authenticated
        try:
            return salt.utils.event.MinionEvent(__opts__, listen=False).fire_event(
                {"data": data, "tag": tag, "events": None, "pretag": None},
                "fire_master",
            )
        except Exception:  # pylint: disable=broad-except
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
            log.debug(lines)
            return False


def fire(data, tag):
    """
    Fire an event on the local minion event bus. Data must be formed as a dict.

    CLI Example:

    .. code-block:: bash

        salt '*' event.fire '{"data":"my event data"}' 'tag'
    """
    try:
        with salt.utils.event.get_event(
            "minion",  # was __opts__['id']
            sock_dir=__opts__["sock_dir"],
            opts=__opts__,
            listen=False,
        ) as event:
            return event.fire_event(data, tag)
    except Exception:  # pylint: disable=broad-except
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        log.debug(lines)
        return False


def send(
    tag,
    data=None,
    preload=None,
    with_env=False,
    with_grains=False,
    with_pillar=False,
    with_env_opts=False,
    **kwargs
):
    """
    Send an event to the Salt Master

    .. versionadded:: 2014.7.0

    :param tag: A tag to give the event.
        Use slashes to create a namespace for related events. E.g.,
        ``myco/build/buildserver1/start``, ``myco/build/buildserver1/success``,
        ``myco/build/buildserver1/failure``.

    :param data: A dictionary of data to send in the event.
        This is free-form. Send any data points that are needed for whoever is
        consuming the event. Arguments on the CLI are interpreted as YAML so
        complex data structures are possible.

    :param with_env: Include environment variables from the current shell
        environment in the event data as ``environ``.. This is a short-hand for
        working with systems that seed the environment with relevant data such
        as Jenkins.
    :type with_env: Specify ``True`` to include all environment variables, or
        specify a list of strings of variable names to include.

    :param with_grains: Include grains from the current minion in the event
        data as ``grains``.
    :type with_grains: Specify ``True`` to include all grains, or specify a
        list of strings of grain names to include.

    :param with_pillar: Include Pillar values from the current minion in the
        event data as ``pillar``. Remember Pillar data is often sensitive data
        so be careful. This is useful for passing ephemeral Pillar values
        through an event. Such as passing the ``pillar={}`` kwarg in
        :py:func:`state.sls <salt.modules.state.sls>` from the Master, through
        an event on the Minion, then back to the Master.
    :type with_pillar: Specify ``True`` to include all Pillar values, or
        specify a list of strings of Pillar keys to include. It is a
        best-practice to only specify a relevant subset of Pillar data.

    :param with_env_opts: Include ``saltenv`` and ``pillarenv`` set on minion
        at the moment when event is send into event data.
    :type with_env_opts: Specify ``True`` to include ``saltenv`` and
        ``pillarenv`` values or ``False`` to omit them.

    :param kwargs: Any additional keyword arguments passed to this function
        will be interpreted as key-value pairs and included in the event data.
        This provides a convenient alternative to YAML for simple values.

    CLI Example:

    .. code-block:: bash

        salt-call event.send myco/mytag foo=Foo bar=Bar
        salt-call event.send 'myco/mytag' '{foo: Foo, bar: Bar}'

    """
    data_dict = {}

    if with_env:
        if isinstance(with_env, list):
            data_dict["environ"] = _dict_subset(with_env, dict(os.environ))
        else:
            data_dict["environ"] = dict(os.environ)

    if with_grains:
        if isinstance(with_grains, list):
            data_dict["grains"] = _dict_subset(with_grains, __grains__)
        else:
            data_dict["grains"] = __grains__.value()

    if with_pillar:
        if isinstance(with_pillar, list):
            data_dict["pillar"] = _dict_subset(with_pillar, __pillar__)
        else:
            data_dict["pillar"] = __pillar__.value()

    if with_env_opts:
        data_dict["saltenv"] = __opts__.get("saltenv", "base")
        data_dict["pillarenv"] = __opts__.get("pillarenv")

    if kwargs:
        data_dict.update(kwargs)

    # Allow values in the ``data`` arg to override any of the above values.
    if isinstance(data, Mapping):
        data_dict.update(data)

    if (
        __opts__.get("local")
        or __opts__.get("file_client") == "local"
        or __opts__.get("master_type") == "disable"
    ) and not __opts__.get("use_master_when_local"):
        return fire(data_dict, tag)
    else:
        return fire_master(data_dict, tag, preload=preload)



import pytest

import salt.modules.cmdmod
import salt.modules.config
import salt.modules.pkg_resource
import salt.modules.yumpkg
import salt.utils.pkg.rpm

pytestmark = [
    pytest.mark.skip_if_binaries_missing("rpm", "yum"),
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(minion_opts, grains):
    grains.update({"osarch": salt.utils.pkg.rpm.get_osarch()})
    return {
        salt.modules.config: {
            "__grains__": grains,
        },
        salt.modules.pkg_resource: {
            "__grains__": grains,
        },
        salt.modules.yumpkg: {
            "__salt__": {
                "cmd.run": salt.modules.cmdmod.run,
                "cmd.run_all": salt.modules.cmdmod.run_all,
                "cmd.run_stdout": salt.modules.cmdmod.run_stdout,
                "config.get": salt.modules.config.get,
                "pkg_resource.add_pkg": salt.modules.pkg_resource.add_pkg,
                "pkg_resource.format_pkg_list": salt.modules.pkg_resource.format_pkg_list,
                "pkg_resource.parse_targets": salt.modules.pkg_resource.parse_targets,
                "pkg_resource.sort_pkglist": salt.modules.pkg_resource.sort_pkglist,
            },
            "__opts__": minion_opts,
            "__grains__": grains,
        },
    }


def test_yum_list_pkgs(grains):
    """
    compare the output of rpm -qa vs the return of yumpkg.list_pkgs,
    make sure that any changes to ympkg.list_pkgs still returns.
    """
    cmd = [
        "rpm",
        "-qa",
        "--queryformat",
        "%{NAME}\n",
    ]
    known_pkgs = salt.modules.cmdmod.run(cmd, python_shell=False)
    listed_pkgs = salt.modules.yumpkg.list_pkgs()
    for line in known_pkgs.splitlines():
        assert any(line in d for d in listed_pkgs)


@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_yumpkg_remove_wildcard():
    salt.modules.yumpkg.install(pkgs=["httpd-devel", "httpd-tools"])
    ret = salt.modules.yumpkg.remove(name="httpd-*")
    assert not ret["httpd-devel"]["new"]
    assert ret["httpd-devel"]["old"]
    assert not ret["httpd-tools"]["new"]
    assert ret["httpd-tools"]["old"]
